import logging

import pygame

from screenManager import screenManager, windowManager
from trosnoth.gui.sound.musicManager import MusicManager
from trosnoth.gui.sound.soundPlayer import SoundPlayer
from trosnoth.utils.utils import timeNow
from trosnoth.gui.errors import ApplicationExit

log = logging.getLogger('app')

TARGET_FRAME_RATE = 60.
LOG_FPS = False


class Application(object):
    '''Instantiating the Main class will set up a ui. Calling the run()
    method will run the application.'''

    def __init__(self, size, fullScreen, caption, element):
        '''Initialise the application.'''
        self._options = (element, size, fullScreen, caption)
        self._initSound()
        self._makeScreenManager(element, size, fullScreen, caption)
        self.fonts = self.screenManager.fonts
        self.initialise()
        self.screenManager.createInterface(element)
        self._running = False
        self.lastTime = None
        self.lastFPSLog = None
        self.ticksSinceFPSLog = 0
        pygame.key.set_repeat(300, 30)

    def __getattr__(self, attr):
        if attr == 'interface':
            return self.screenManager.interface
        raise AttributeError(attr)

    def _initSound(self):
        self.musicManager = MusicManager()
        self.soundPlayer = SoundPlayer()

    def _makeScreenManager(self, element, size, fullScreen, caption):
        self.screenManager = screenManager.ScreenManager(
            self, size, fullScreen, caption)

    def restart(self):
        self._initSound()
        self._makeScreenManager(*self._options)
        self.fonts = self.screenManager.fonts
        self.initialise()
        self.screenManager.createInterface(self._options[0])

    def initialise(self):
        '''
        Provides the opportunity for initialisation by subclasses before the
        interface element is created.
        '''

    def getFontFilename(self, fontName):
        '''
        May be overridden by an application in order to provide a custom font
        location, or some other mapping from font name to filename.
        '''
        return fontName

    def run(self):
        '''Runs the application.'''
        def _stop():
            self._running = False
            raise ApplicationExit
        self._stop = _stop

        self._running = True
        self.lastTime = self.lastFPSLog = timeNow()
        while self._running:
            try:
                self.tick()
                self.doFlip()
            except ApplicationExit:
                break
        pygame.quit()

    def run_twisted(self, reactor=None):
        '''Runs the application using Twisted's reactor.'''

        if reactor is None:
            from twisted.internet import reactor

        def _stop():
            if not self._running:
                if reactor.running:
                    log.warning(
                        'stop() called twice. Terminating immediately.')
                    reactor.stop()
                return

            self._running = False

            # Give time for shutdown sequence.
            reactor.callLater(0.3, reactor.stop)

        self._stop = _stop
        self.lastTime = self.lastFPSLog = timeNow()

        self.targetFrameInterval = 1. / TARGET_FRAME_RATE
        reactor.callLater(self.targetFrameInterval, self.twisted_tick, reactor)
        self._running = True
        reactor.run()
        pygame.display.quit()
        pygame.quit()

    def run_with_profiling(self, twisted, *args, **kwargs):
        import cProfile
        from trosnoth.utils.profiling import KCacheGrindOutputter
        prof = cProfile.Profile()

        if twisted:
            mainFunction = self.run_twisted
        else:
            mainFunction = self.run

        try:
            prof.runcall(mainFunction, *args, **kwargs)
        except SystemExit:
            pass
        finally:
            kg = KCacheGrindOutputter(prof)
            with open('trosnoth.log', 'wb') as f:
                kg.output(f)

    def twisted_tick(self, reactor):
        reactor.callLater(self.targetFrameInterval, self.twisted_tick, reactor)
        self.tick()
        self.doFlip()

    def stop(self):
        if self._running:
            self.stopping()
        self._stop()

    def tick(self):
        try:
            self._tick()
        except ApplicationExit:
            raise
        except Exception, e:
            log.exception(str(e))

    def _tick(self):
        '''Processes the events in the pygame event queue, and causes the
        application to be updated, then refreshes the screen. This routine is
        called as often as possible - it is not limited to a specific frame
        rate.'''
        if not self._running:
            return

        now = timeNow()

        self.ticksSinceFPSLog += 1
        if LOG_FPS and now - self.lastFPSLog > 1:
            log.warning('%.2f FPS', self.ticksSinceFPSLog / (
                now - self.lastFPSLog))
            self.lastFPSLog = now
            self.ticksSinceFPSLog = 0

        # Process the events in the event queue.
        for event in pygame.event.get():
            try:
                event = self.musicManager.processEvent(event)
                if not self._running:
                    return
                if event is not None:
                    event = self.screenManager.processEvent(event)
                    if not self._running:
                        return
                    if event is not None:
                        # Special events.
                        if event.type == pygame.QUIT:
                            self.stop()
                            return
            except ApplicationExit:
                raise
            except Exception, e:
                log.exception(str(e))
            if not self._running:
                return

        # Give things an opportunity to update their state.
        deltaT = now - self.lastTime
        self.lastTime = now
        try:
            self.screenManager.tick(deltaT)
        except ApplicationExit:
            raise
        except Exception, e:
            log.error('UNHANDLED ERROR IN TICK')
            log.exception(str(e))

        if not self._running:
            return

        if pygame.display.get_active():
            # Update the screen.
            try:
                self.screenManager.draw(self.screenManager.screen)
            except pygame.error, e:
                # Surface may have been lost (DirectDraw error)
                log.exception('Pygame error: %s' % (e,))

    def doFlip(self):
        if self._running and pygame.display.get_active():
            # Flip the display.
            pygame.display.flip()

    def stopping(self):
        '''Any finalisation which must happen before stopping the reactor.'''
        self.screenManager.finalise()


class MultiWindowApplication(Application):

    def _makeScreenManager(self, element, size, fullScreen, caption):
        self.screenManager = windowManager.WindowManager(
            self, element, size, fullScreen, caption)



def collideTrajectory(pt, origin, trajectory, tolerance):
    '''Returns true if pt lies within a distance of tolerance from the
    line segment described by origin and trajectory.'''

    # Get distance between point and trajectory origin.
    tr0 = origin
    d0 = ((tr0[0] - pt[0]) ** 2 + (tr0[1] - pt[1]) ** 2) ** 0.5
    if d0 < tolerance:
        return True

    delta = trajectory

    while True:
        # Calculate other end of trajectory.
        tr1 = (tr0[0] + delta[0],
               tr0[1] + delta[1])
        d1 = ((tr1[0] - pt[0]) ** 2 + (tr1[1] - pt[1]) ** 2) ** 0.5

        # Check for collision.
        if d1 < tolerance:
            return True

        # Refine and loop.
        if d1 < d0:
            tr0, delta = tr1, (-0.5 * delta[0],
                               -0.5 * delta[1])
        else:
            delta = (0.5 * delta[0],
                     0.5 * delta[1])

        # Check end condition.
        if (delta[0] ** 2 + delta[1] ** 2) < 5:
            return False



from trosnoth.gui.framework import framework
import pygame


class ViewControlInterface(framework.Element):
    '''Interface for controlling the replay view.'''

    # The virtual keys we care about.
    state_vkeys = frozenset(['left', 'right', 'jump', 'down'])

    def __init__(self, app, gameInterface):
        super(ViewControlInterface, self).__init__(app)

        world = gameInterface.world
        self.gameInterface = gameInterface
        self.keyMapping = gameInterface.keyMapping

        self.world = world
        self._state = dict([(k, False) for k in self.state_vkeys])

        self.vx = 0
        self.vy = 0

    def updateState(self, state, enabled):
        self._state[state] = enabled
        if self._state['left'] and not self._state['right']:
            self.vx = -1000
        elif self._state['right'] and not self._state['left']:
            self.vx = 1000
        else:
            self.vx = 0

        if self._state['jump'] and not self._state['down']:
            self.vy = -1000
        elif self._state['down'] and not self._state['jump']:
            self.vy = 1000
        else:
            self.vy = 0

    def tick(self, deltaT):
        if self.vx != 0 or self.vy != 0:
            x, y = self.gameInterface.gameViewer.viewManager.getTargetPoint()
            x += self.vx * deltaT
            y += self.vy * deltaT
            self.gameInterface.gameViewer.setTarget((x, y))

    def processEvent(self, event):
        '''
        Event processing works in the following way:
        1. If there is a prompt on screen, the prompt will either use the
        event, or pass it on.
        2. If passed on, the event will be sent back to the main class, for it
        to process whether player movement uses this event. If it doesn't use
        the event, it will pass it back.
        3. If so, the hotkey manager will see if the event means anything to
        it. If not, that's the end, the event is ignored.
        '''

        # Handle events specific to in-game.
        if event.type == pygame.KEYDOWN:
            try:
                stateKey = self.keyMapping[event.key]
            except KeyError:
                return event

            if stateKey not in self.state_vkeys:
                return event

            self.updateState(stateKey, True)
        elif event.type == pygame.KEYUP:
            try:
                stateKey = self.keyMapping[event.key]
            except KeyError:
                return event

            if stateKey not in self.state_vkeys:
                return event

            self.updateState(stateKey, False)
        else:
            return event



import logging
import pygame
import random

from trosnoth.model.obstacles import (
    GroundObstacle, LedgeObstacle, VerticalWall, RoofObstacle, joinObstacles)
from trosnoth.model.map import MapLayout

log = logging.getLogger(__name__)


class MapBlockDef(object):
    '''Represents the static information about a particular map block. A map
    block is a grid square, and may contain a single zone, or the interface
    between two zones.'''

    def __init__(self, kind, x, y):
        self.pos = (x, y)       # Pos is the top-left corner of this block.
        assert kind in ('top', 'btm', 'fwd', 'bck')
        self.kind = kind
        self.layout = None

        self.obstacles = []
        self.ledges = []
        self.indices = MapLayout.getMapBlockIndices(x, y)

        self.rect = pygame.Rect(x, y, 0, 0)
        self.rect.size = (self._getWidth(), MapLayout.halfZoneHeight)

        self.graphics = None
        self.blocked = False    # There's a barrier somewhere depending on type

        self.nextObstacleId = 0

        # For body block.
        self.zone = None

        # For interface block.
        self.zone1 = None
        self.zone2 = None

    def __repr__(self):
        return 'MapBlockDef(%r @ %r)' % (self.kind, self.indices)

    def __str__(self):
        return 'Block @ %r' % (self.indices,)

    def _getWidth(self):
        if self.kind in ('top', 'btm'):
            return MapLayout.zoneBodyWidth
        else:
            return MapLayout.zoneInterfaceWidth

    def _getHeight(self):
        return MapLayout.halfZoneHeight

    def getObstacleById(self, obstacleId):
        if obstacleId >= len(self.obstacles):
            return self.ledges[obstacleId - len(self.obstacles)]
        return self.obstacles[obstacleId]

    def spawnState(self, universe, zoneWithDef):
        if self.kind == 'top':
            return TopBodyMapBlock(
                universe, self, zoneWithDef.get(self.zone, None))
        elif self.kind == 'btm':
            return BottomBodyMapBlock(
                universe, self, zoneWithDef.get(self.zone, None))
        elif self.kind == 'fwd':
            return ForwardInterfaceMapBlock(
                universe, self,
                zoneWithDef.get(self.zone1, None),
                zoneWithDef.get(self.zone2, None))
        elif self.kind == 'bck':
            return BackwardInterfaceMapBlock(
                universe, self,
                zoneWithDef.get(self.zone1, None),
                zoneWithDef.get(self.zone2, None))
        else:
            assert False

    def randomPosition(self):
        '''
        Returns a random position within this map block.
        '''
        x = self.rect.left + random.random() * self.rect.width
        y = self.rect.top + random.random() * self.rect.height
        return x, y

    def randomPositionInZone(self, zone):
        '''
        Returns a random position within this map block and within the given
        zone.
        '''
        if zone not in self.getZones():
            raise ValueError('Invalid zone!')
        while True:
            x, y = self.randomPosition()
            if self.getZoneAtPoint(x, y) == zone:
                return x, y

    def getZones(self):
        result = []
        for z in (self.zone, self.zone1, self.zone2):
            if z is not None:
                result.append(z)
        return result

    def getZoneAtPoint(self, x, y):
        '''getZoneAtPoint(x, y)
        Returns the zone def for the zone at the specified point, ASSUMING
        that the point is in fact within this map block.'''
        if self.kind == 'fwd':
            # Equation of interface line:
            #   (y - self.y) = -(halfZoneHeight / interfaceWidth)(x - self.x)
            #                        + halfZoneHeight
            deltaY = y - self.pos[1] - MapLayout.halfZoneHeight
            deltaX = x - self.pos[0]

            if (deltaY * MapLayout.zoneInterfaceWidth >
                    -MapLayout.halfZoneHeight * deltaX):
                return self.zone2
            return self.zone1
        elif self.kind == 'bck':
            # Equation of interface line:
            #   (y - self.y) = (halfZoneHeight / interfaceWidth)(x - self.x)
            deltaY = y - self.pos[1]
            deltaX = x - self.pos[0]

            if (deltaY * MapLayout.zoneInterfaceWidth >
                    MapLayout.halfZoneHeight * deltaX):
                return self.zone1
            return self.zone2
        else:
            return self.zone

    def addPlatform(self, pos, dx, reverse):
        '''Adds a horizontal platform which can be dropped through to this
        block's list of obstacles.

        pos         The position of the obstacle's first point relative to the
                    top-left corner of this block, or relative to the top-right
                    corner if reverse is set to True.
        dx          The horizontal displacement from first point to the second
                    point of the obstacle. This value should always be
                    positive.
        reverse     Determines whether the obstacle is defined in terms of the
                    top-left corner of this map block or the top-right corner.
        '''

        # Add the block's position offset.
        pt = [pos[0] + self.pos[0],
              pos[1] + self.pos[1]]
        if reverse:
            pt[0] += self.rect.width

        # Put the platform in.
        x, y = pt

        ledge = LedgeObstacle(pt, (dx, 0))
        ledge.setId(self.nextObstacleId)
        self.ledges.append(ledge)
        assert self.getObstacleById(self.nextObstacleId) is ledge
        self.nextObstacleId += 1

    def addObstacle(self, points, reverse):
        '''Adds an obstacle to this block's list of obstacles.

        points      A sequence specifying the positions of the obstacle's
                    points relative to the top-left corner of this block, or
                    relative to the top-right corner if reverse is set to True.
        reverse     Determines whether the obstacle is defined in terms of the
                    top-left corner of this map block or the top-right corner.
        '''
        resultingObstacles = []

        def translate(pt):
            '''
            Translates the give point to absolute reference points given that
            it's relative to this map block.
            '''
            x = pt[0] + self.pos[0]
            y = pt[1] + self.pos[1]
            if reverse:
                x += self.rect.width
            return x, y

        lastGround = [None]

        def checkGround(obstacle):
            if obstacle.deltaPt[0] > 0:
                if lastGround[0]:
                    obstacle.leftGround = lastGround[0]
                    lastGround[0].rightGround = obstacle
                lastGround[0] = obstacle
            else:
                lastGround[0] = None

        def makeObstacle(pt1, pt2):
            '''
            Creates, records and returns a new obstacle between the two points
            provided.
            '''
            dx, dy = pt2[0] - pt1[0], pt2[1] - pt1[1]

            if dx == 0:
                kind = VerticalWall
            elif dx > 0:
                kind = GroundObstacle
            else:
                kind = RoofObstacle

            return kind(pt1, (dx, dy))

        firstObstacle = None
        prevObstacle = None
        prevPt = None
        finalObstacles = []
        for pt in points:
            pt = translate(pt)
            if prevPt is not None:
                obstacle = makeObstacle(prevPt, pt)
                if prevObstacle is None:
                    firstObstacle = obstacle
                else:
                    finalObstacles.extend(
                        joinObstacles(prevObstacle, obstacle))
                resultingObstacles.append(obstacle)
                prevObstacle = obstacle
            prevPt = pt
        if points[0] == points[-1] and firstObstacle is not None:
            if reverse:
                finalObstacles[:0] = joinObstacles(obstacle, firstObstacle)
            else:
                finalObstacles.extend(joinObstacles(obstacle, firstObstacle))

        if reverse:
            resultingObstacles.reverse()
            finalObstacles.reverse()

        for obstacle in resultingObstacles + finalObstacles:
            obstacle.setId(self.nextObstacleId)
            self.obstacles.append(obstacle)
            assert self.getObstacleById(self.nextObstacleId) is obstacle
            self.nextObstacleId += 1

    def coalesceLayout(self):
        '''
        Gives this block definition the chance to perform calculations after
        all obstacles have been added to it.
        '''
        touched = set()
        for obs in self.obstacles:
            if obs in touched or not obs.ground:
                continue
            subshape = set()
            pending = set([obs])
            while len(pending) > 0:
                obs = pending.pop()
                if obs in subshape:
                    continue
                subshape.add(obs)
                touched.add(obs)
                if obs.leftGround is not None:
                    pending.add(obs.leftGround)
                if obs.rightGround is not None:
                    pending.add(obs.rightGround)
            for obs in subshape:
                obs.subshape = subshape


class MapBlock(object):
    '''
    Represents a grid square of the map which may contain a single zone,
    or the interface between two zones.
    '''

    def __init__(self, universe, defn):
        self.universe = universe
        self.defn = defn

    def __contains__(self, point):
        '''Checks whether a given point is within this zone.'''
        return self.defn.rect.collidepoint(*point)

    def _getBlockLeft(self):
        try:
            i, j = self.defn.indices
            if j == 0:
                return None
            return self.universe.zoneBlocks[i][j - 1]
        except IndexError:
            return None
    blockLeft = property(_getBlockLeft)

    def _getBlockRight(self):
        try:
            i, j = self.defn.indices
            if j >= len(self.universe.zoneBlocks[i]) - 1:
                return None
            return self.universe.zoneBlocks[i][j + 1]
        except IndexError:
            return None
    blockRight = property(_getBlockRight)

    def _getBlockAbove(self):
        try:
            i, j = self.defn.indices
            if i == 0:
                return None
            return self.universe.zoneBlocks[i - 1][j]
        except IndexError:
            return None
    blockAbove = property(_getBlockAbove)

    def _getBlockBelow(self):
        try:
            i, j = self.defn.indices
            if i >= len(self.universe.zoneBlocks) - 1:
                return None
            return self.universe.zoneBlocks[i + 1][j]
        except IndexError:
            return None
    blockBelow = property(_getBlockBelow)

    def getZoneAtPoint(self, x, y):
        '''getZoneAtPoint(x, y)
        Returns what zone is at the specified point, ASSUMING that the point
        is in fact within this map block.'''
        raise NotImplementedError('getZoneAtPoint not implemented.')

    def fromEdge(self, player):
        '''Returns the distance from the edge of the zone that a player is'''
        return


class BodyMapBlock(MapBlock):
    '''Represents a map block which contains only a single zone.'''

    def __init__(self, universe, defn, zone):
        super(BodyMapBlock, self).__init__(universe, defn)
        self.zone = zone

    def __str__(self):
        return '< %s >' % self.zone

    def getZoneAtPoint(self, x, y):
        return self.zone

    def getZones(self):
        if self.zone is not None:
            yield self.zone


class TopBodyMapBlock(BodyMapBlock):
    '''Represents a map block containing the top half of a zone.'''

    def orbPos(self, drawRect):
        return drawRect.midbottom

    def fromEdge(self, player):
        '''Returns the distance from the edge of the zone that a player is'''

        return min(
            abs(player.pos[1] - self.defn.rect.top),
            self.blockLeft.fromEdge(player),
            self.blockRight.fromEdge(player))


class BottomBodyMapBlock(BodyMapBlock):
    '''Represents a map block containing the bottom half of a zone.'''

    def orbPos(self, drawRect):
        return drawRect.midtop

    def fromEdge(self, player):
        '''Returns the distance from the edge of the zone that a player is'''

        return min(
            abs(self.defn.rect.bottom - player.pos[1]),
            self.blockLeft.fromEdge(player),
            self.blockRight.fromEdge(player))


class InterfaceMapBlock(MapBlock):
    '''Base class for map blocks which contain the interface between two
    zones.'''

    def __init__(self, universe, defn, zone1, zone2):
        super(InterfaceMapBlock, self).__init__(universe, defn)

        self.zone1 = zone1
        self.zone2 = zone2

    def getZones(self):
        if self.zone1 is not None:
            yield self.zone1
        if self.zone2 is not None:
            yield self.zone2


class ForwardInterfaceMapBlock(InterfaceMapBlock):
    '''Represents a map block which contains the interface between two
    zones, split in the direction of a forward slash '/'.
    Note that exactly on the diagonal counts as being in the left-hand zone.
    '''

    def __str__(self):
        return '< %s / %s >' % (self.zone1, self.zone2)

    def getZoneAtPoint(self, x, y):
        # Equation of interface line:
        #   (y - self.y) = -(halfZoneHeight / interfaceWidth)(x - self.x)
        #                        + halfZoneHeight
        deltaY = y - self.defn.pos[1] - MapLayout.halfZoneHeight
        deltaX = x - self.defn.pos[0]

        if (deltaY * MapLayout.zoneInterfaceWidth > -MapLayout.halfZoneHeight *
                deltaX):
            return self.zone2
        return self.zone1

    def fromEdge(self, player):
        '''Returns the distance from the edge of the zone that a player is.
        Assumes that the player is actually in this map block.'''
        relPos = (
            self.defn.rect.left - player.pos[0],
            self.defn.rect.top - player.pos[1])

        # Note: this following formula relies upon the dimensions:
        # MapLayout.halfZoneHeight = 384
        # MapLayout.zoneInterfaceWidth = 512
        # Should they be changed, should change this to:
        # d = sin(theta) * abs(relPos[1] + relPos[0] *
        # MapLayout.halfZoneHeight / MapLayout.zoneInterfaceWidth + 384)
        # (where theta is the angle formed by the diagonal line separating the
        # zones, and a vertical line).
        d = 0.8 * abs(relPos[1] + relPos[0] * 0.75 + 384)
        return d


class BackwardInterfaceMapBlock(InterfaceMapBlock):
    '''Represents a map block which contains the interface between two
    zones, split in the direction of a backslash '\'.
    Note that a point exactly on the diagonal counts as being in the left-hand
    zone.
    '''

    def __str__(self):
        return '< %s \ %s >' % (self.zone1, self.zone2)

    def getZoneAtPoint(self, x, y):
        # Equation of interface line:
        #   (y - self.y) = (halfZoneHeight / interfaceWidth)(x - self.x)
        deltaY = y - self.defn.pos[1]
        deltaX = x - self.defn.pos[0]

        if (
                deltaY * MapLayout.zoneInterfaceWidth >
                MapLayout.halfZoneHeight * deltaX):
            return self.zone1
        return self.zone2

    def fromEdge(self, player):
        '''Returns the distance from the edge of the zone that a player is.
        Assumes that the player is actually in this map block.'''
        relPos = (
            self.defn.rect.left - player.pos[0],
            self.defn.rect.top - player.pos[1])

        # Note: this following formula relies upon the dimensions:
        # MapLayout.halfZoneHeight = 384
        # MapLayout.zoneInterfaceWidth = 512
        # Should they be changed, should change this to:
        # d = sin(theta) * abs(relPos[1] - relPos[0] *
        # MapLayout.halfZoneHeight / MapLayout.zoneInterfaceWidth)
        # where theta is the angle formed by the diagonal line separating the
        # zones, and a vertical line.
        d = 0.8 * abs(relPos[1] - relPos[0] * 3 / 4)
        return d



'''
modes.py: defines a set of pre-defined game modes
'''

import logging

log = logging.getLogger(__name__)


class PhysicsConstants(object):
    STATE_ATTRS = [
        'normal',
        'shooting',
        'shotSpeed',
        'shotLifetime',
        'fractionOfPlayerVelocityImpartedToShots',
        'playerXVel',
        'playerSlowXVel',
        'playerMaxGhostVel',
        'playerJumpThrust',
        'playerMaxFallVel',
        'playerMaxJumpTime',
        'playerGravity',
        'playerOwnDarkReloadTime',
        'playerNeutralReloadTime',
        'playerEnemyDarkReloadTime',
        'playerTurretReloadTime',
        'playerMachineGunReloadTime',
        'playerMachineGunFireRate',
        'playerTurretHeatCapacity',
        'playerShotHeat',
        'playerRespawnTotal',
        'playerBounce',
        'playerRespawnHealth',
        'playerShoxwaveReloadTime',
        'grenadeMaxFallVel',
        'grenadeGravity',
        'grenadeInitVel',
        'trosballThrowVel',
        'trosballExplodeTime',
    ]

    def __init__(self):
        self.setModeNormal()

    def hasMode(self, gameMode):
        return hasattr(self, 'setMode' + gameMode)

    def setMode(self, gameMode):
        try:
            fn = getattr(self, 'setMode' + gameMode)
        except (AttributeError, TypeError):
            return False

        fn()
        return True

    def _standard(self, pace=1, fireRate=1, gravity=1, jumpHeight=1,
            respawnRate=1, shotSpeed=1, shotLength=1, bounce=False,
            respawnHealth=1, shooting=True, trosballTime=5, shotFactor=1,
            normal=False):

        self.normal = normal
        self.shooting = shooting

        # Speed that shots travel at.
        self.shotSpeed = 600 * shotSpeed                      # pix/s
        self.shotLifetime = (1. / shotSpeed) * shotLength     # s
        self.fractionOfPlayerVelocityImpartedToShots = 0.3 * shotFactor

        # The following values control player movement.
        self.playerXVel = 360 * pace                  # pix/s
        self.playerSlowXVel = 180 * pace              # pix/s
        self.playerMaxGhostVel = 675. * pace          # pix/s
        self.playerJumpThrust = 540 * jumpHeight      # pix/s
        self.playerMaxFallVel = 540. * gravity        # pix/s
        self.playerMaxJumpTime = 0.278                # s
        self.playerGravity = 3672 * gravity           # pix/s/s
        self.playerOwnDarkReloadTime = 1 / 2.7 / fireRate      # s
        self.playerNeutralReloadTime = 1 / 2. / fireRate   # s
        self.playerEnemyDarkReloadTime = 1 / 1.4 / fireRate    # s
        self.playerTurretReloadTime = 0.083 / fireRate     # s
        self.playerMachineGunReloadTime = 4 / fireRate   # s
        self.playerMachineGunFireRate = 0.1 / fireRate   # s
        self.playerTurretHeatCapacity = 2.4 * fireRate
        self.playerShotHeat = 0.4
        self.playerRespawnTotal = 7.5 / respawnRate
        self.playerBounce = bounce
        self.playerRespawnHealth = respawnHealth
        self.playerShoxwaveReloadTime = self.playerEnemyDarkReloadTime

        self.grenadeMaxFallVel = 540. * gravity
        self.grenadeGravity = 300 * gravity
        self.grenadeInitVel = 400 * pace
        self.trosballThrowVel = 1000 * pace

        self.trosballExplodeTime = trosballTime

        # For calculation of cost of shortest path
        xMax = max(self.playerXVel, self.playerSlowXVel)
        yMax = max(self.playerMaxFallVel, self.playerJumpThrust)
        self.maxLivePlayerSpeed = (xMax ** 2 + yMax ** 2) ** 0.5

    def dumpState(self):
        return {k: getattr(self, k) for k in self.STATE_ATTRS}

    def restoreState(self, data):
        for k, v in data.iteritems():
            if k in self.STATE_ATTRS:
                setattr(self, k, v)

    def setModeNoShots(self):
        self._standard(shooting=False)

    def setModeLightning(self):
        self._standard(pace=1.75, fireRate=2)

    def setModeLowGravity(self):
        self._standard(gravity=0.25)

    def setModeNormal(self):
        self._standard(normal=True)

    def setModeFastFire(self):
        self._standard(fireRate=3, shotSpeed=2)

    def setModeInsane(self):
        self._standard(pace=1.75, fireRate=3, jumpHeight=2, shotSpeed=2,
                        shotLength=2, respawnRate=60)

    def setModeFastRespawn(self):
        self._standard(respawnRate=60)

    def setModeSlow(self):
        self._standard(pace=0.5, gravity=0.25, jumpHeight=0.7, shotSpeed=0.5)

    def setModeHighFastFall(self):
        self._standard(jumpHeight=2, gravity=2)

    def setModeLaser(self):
        self._standard(shotSpeed=30, shotLength=100)

    def setModeManyShots(self):
        self._standard(shotLength=20)

    def setModeZeroG(self):
        self._standard(gravity=0.001)

    def setModeAntiG(self):
        self._standard(gravity=-0.1)

    def setModeBouncy(self):
        self._standard(bounce=True)

    def setModeHighBouncy(self):
        self._standard(bounce=True, jumpHeight=3)

    def setModeTwoLives(self):
        self._standard(respawnHealth=2)

    def setRapidTrosball(self):
        self._standard(trosballTime=1)



from django import forms
from django.core import urlresolvers
from django.contrib import admin
from django.db import models

from trosnoth.djangoapp.models import (
    TrosnothUser, AchievementProgress, GameRecord, GamePlayer, PlayerKills,
    TrosnothServerSettings, UpgradesUsedInGameRecord,
)

DEFAULT_OVERRIDES = {
    models.TextField: {'widget': admin.widgets.AdminTextInputWidget},
}


@admin.register(TrosnothServerSettings)
class TrosnothServerSettingsAdmin(admin.ModelAdmin):
    formfield_overrides = DEFAULT_OVERRIDES

    def has_add_permission(self, request):
        # if there's already an entry, do not allow adding
        count = TrosnothServerSettings.objects.all().count()
        if count == 0:
            return True

        return False


class AchievementProgressInline(admin.TabularInline):
    model = AchievementProgress
    formfield_overrides = DEFAULT_OVERRIDES


@admin.register(TrosnothUser)
class TrosnothUserAdmin(admin.ModelAdmin):
    inlines = [
        AchievementProgressInline,
    ]


class GamePlayerInline(admin.TabularInline):
    model = GamePlayer
    formfield_overrides = DEFAULT_OVERRIDES
    fields = ('user', 'botName', 'team', 'detailLink')
    readonly_fields = ('detailLink',)

    def detailLink(self, instance):
        if not instance.id:
            return u''

        changeform_url = urlresolvers.reverse(
            'admin:trosnoth_gameplayer_change', args=(instance.id,)
        )
        return u'<a href="%s">Details</a>' % changeform_url
    detailLink.allow_tags = True
    detailLink.short_description = ''


@admin.register(GameRecord)
class GameRecordAdmin(admin.ModelAdmin):
    formfield_overrides = DEFAULT_OVERRIDES
    inlines = [
        GamePlayerInline,
    ]


class UpgradesUsedInGameRecordAdmin(admin.TabularInline):
    model = UpgradesUsedInGameRecord
    formfield_overrides = DEFAULT_OVERRIDES


class PlayerKillsInline(admin.TabularInline):
    model = PlayerKills
    fk_name = 'killee'
    formfield_overrides = DEFAULT_OVERRIDES


@admin.register(GamePlayer)
class GamePlayerAdmin(admin.ModelAdmin):
    model = GamePlayer
    formfield_overrides = DEFAULT_OVERRIDES
    inlines = [
        UpgradesUsedInGameRecordAdmin,
        PlayerKillsInline,
    ]



import logging
import platform

import pygame

from trosnoth.gui.framework import framework
from trosnoth.gui.common import Location
from trosnoth.utils.event import Event

log = logging.getLogger('screenManager')


class SystemHotKeys(framework.Element):
    '''
    Traps a Ctrl+Break / Ctrl+Pause keypress and terminates the game.
    Traps an Alt+Tab keypress and minimises the game.
    '''
    def processEvent(self, event):
        if event.type == pygame.KEYDOWN:
            if event.mod & pygame.KMOD_CTRL and event.key in (
                    pygame.K_SCROLLOCK, pygame.K_BREAK, pygame.K_PAUSE):
                log.warning('Trapped Ctrl+Break keypress: Quitting.')
                self.app.stop()
                return None
            if (
                    event.key == pygame.K_BACKSLASH
                    and event.mod & pygame.KMOD_CTRL
                    and event.mod & pygame.KMOD_META):
                log.warning('Trapped Ctrl+Meta+Backslash: Quitting.')
                self.app.stop()
                return None
            if (event.key == pygame.K_TAB and (event.mod & pygame.KMOD_ALT) and
                    self.app.screenManager.isFullScreen()):
                if platform.system() != 'Darwin':
                    # Minimise.
                    pygame.display.iconify()
                return None
        return event


class Pointer(framework.CompoundElement):
    def __init__(self, app, element):
        super(Pointer, self).__init__(app)
        assert hasattr(element, 'setPos')
        self.elements = [element]

    def draw(self, screen):
        try:
            pos = pygame.mouse.get_pos()
            self.elements[0].setPos(Location(pos, 'center'))
        except Exception, e:
            log.exception(str(e))
        super(Pointer, self).draw(screen)


class ScreenFonts(object):
    def addFont(self, identifier, font):
        setattr(self, identifier, font)


class ScreenManager(framework.CompoundElement):
    '''Manages everything to do with the screen, especially any scaling due
    to the size of the screen.'''

    def __init__(self, app, size, fullScreen, caption):
        super(ScreenManager, self).__init__(app)
        pygame.init()

        self.onResize = Event()
        self.size = None
        self.setScreenProperties(size, fullScreen, caption)
        self.fonts = ScreenFonts()
        self.pointer = None
        self.interface = None
        self.terminator = SystemHotKeys(self.app)
        self.elements = [self.terminator]

    def processEvent(self, event):
        if event.type == pygame.VIDEORESIZE:
            self.processResize(event.size)
            return None
        return super(ScreenManager, self).processEvent(event)

    def setPointer(self, pointer):
        # Set up a sprite for the mouse pointer.
        self.elements.remove(self.terminator)
        if self.pointer is not None and self.pointer in self.elements:
            self.elements.remove(self.pointer)
        self.pointer = Pointer(self.app, pointer)
        self.elements.append(self.pointer)
        self.elements.append(self.terminator)

    def createInterface(self, element):
        self.interface = element(self.app)
        self.elements = [self.interface, self.terminator]
        if self.pointer is not None:
            self.elements.append(self.pointer)
        self.elements.append(self.terminator)

    def setScreenProperties(self, size, fullScreen, caption):
        pygame.display.set_caption(caption)
        if fullScreen:
            self.settings = pygame.FULLSCREEN
        else:
            self.settings = pygame.RESIZABLE
        self.settings |= pygame.HWSURFACE | pygame.DOUBLEBUF

        self.processResize(size)

    def processResize(self, size):
        settings = self.settings

        try:
            self.screen = pygame.display.set_mode(size, settings)
        except pygame.error:
            # Resolution not valid for this display: get the highest
            # valid resolution and use that instead
            resolutionList = pygame.display.list_modes()
            size = resolutionList[0]
            self.screen = pygame.display.set_mode(size, settings)

        if size != self.size:
            self.size = size
            self.scaleFactor = min(size[0] / 1024.,
                                   size[1] / 768.)

            # Calculate corner offset for strange screen sizes.
            self.scaledSize = self.scale((1024, 768))
            self.offsets = ((size[0] - self.scaledSize[0]) / 2,
                            (size[1] - self.scaledSize[1]) / 2)

            self.rect = self.screen.get_rect()
            self.scaledRect = pygame.Rect(self.offsets, self.scaledSize)

            self.onResize()

    def finalise(self):
        pygame.mouse.set_visible(1)

    def isFullScreen(self):
        return (self.settings & pygame.FULLSCREEN) != 0

    def scale(self, point):
        '''Takes a point on a 1024x768 screen and scales it to this screen.'''
        return (int(point[0] * self.scaleFactor),
                int(point[1] * self.scaleFactor))

    def placePoint(self, point):
        '''Places the point from a 1024x768 screen onto the largest rectangle
        of that ratio that fits within this screen.'''
        return (int(point[0] * self.scaleFactor + self.offsets[0]),
                int(point[1] * self.scaleFactor + self.offsets[1]))



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging

from twisted.internet import reactor
from twisted.internet.error import CannotListenError

from trosnoth.model.agenthub import LocalHub
from trosnoth.network.base import MsgServer
from trosnoth.network.networkDefines import serverVersion

from trosnoth.messages import (
    ChatMsg, InitClientMsg,
    AbandonUpgradeMsg, ResyncAcknowledgedMsg, BuyUpgradeMsg, ShootMsg,
    RespawnRequestMsg, JoinRequestMsg, UpdatePlayerStateMsg, AimPlayerAtMsg,
    PlayerIsReadyMsg, SetPreferredDurationMsg, SetPreferredTeamMsg,
    SetPreferredSizeMsg, RemovePlayerMsg, ChangeNicknameMsg, CheckSyncMsg,
    ThrowTrosballMsg, PlayerHasUpgradeMsg,
)
from trosnoth.utils import netmsg
from trosnoth.utils.message import UnhandledMessage
from trosnoth.utils.event import Event

log = logging.getLogger('server')

# The set of messages that the server expects to receive.
serverMsgs = netmsg.MessageCollection(
    ShootMsg,
    AbandonUpgradeMsg,
    UpdatePlayerStateMsg,
    AimPlayerAtMsg,
    BuyUpgradeMsg,
    ThrowTrosballMsg,
    RespawnRequestMsg,
    JoinRequestMsg,
    ChatMsg,
    PlayerIsReadyMsg,
    SetPreferredDurationMsg,
    SetPreferredTeamMsg,
    SetPreferredSizeMsg,
    RemovePlayerMsg,
    ChangeNicknameMsg,
    PlayerHasUpgradeMsg,
    CheckSyncMsg,
    ResyncAcknowledgedMsg,
)


class TrosnothServerFactory(MsgServer):
    messages = serverMsgs

    def __init__(self, game, authTagManager=None, *args, **kwargs):
        self.game = game
        self.authTagManager = authTagManager

        self.connectedClients = set()

        self.onShutdown = Event()       # ()

        self.running = True
        self._alreadyShutdown = False

    def checkGreeting(self, greeting):
        return (greeting == 'Trosnoth18')

    def startListening(self, port=6789, interface=''):
        try:
            self.port = reactor.listenTCP(port, self, interface=interface)
        except CannotListenError:
            log.warning('WARNING: Could not listen on port %s', port)
            self.port = reactor.listenTCP(0, self, interface=interface)

    def getTCPPort(self):
        return self.port.getHost().port

    def stopListening(self):
        self.port.stopListening()

    def gotBadString(self, protocol, data):
        log.warning('Server: Unrecognised network data: %r' % (data,))
        log.warning('      : Did you invent a new network message and forget')
        log.warning('      : to add it to trosnoth.network.server.serverMsgs?')

    def connectionEstablished(self, protocol):
        '''
        Called by the network manager when a new incoming connection is
        completed.
        '''
        # Remember that this connection's ready for transmission.
        self.connectedClients.add(protocol)
        hub = LocalHub(self.game)
        hub.connectNode(protocol)

        # Send the setting information.
        protocol.gotServerCommand(InitClientMsg(self._getClientSettings()))

    def _sendBackToRequestPlug(self, msg):
        raise UnhandledMessage(
            'Server will not send %r backwards over ' 'request plug' % (msg,))

    def _getClientSettings(self):
        '''Returns a string representing the settings which must be sent to
        clients that connect to this server.'''

        result = self.game.world.dumpEverything()
        result['serverVersion'] = serverVersion

        return repr(result)

    def connectionLost(self, protocol, reason):
        if protocol in self.connectedClients:
            protocol.hub.disconnectNode()

        self.connectedClients.remove(protocol)

        # Check for game over and no connections left.
        if (len(self.connectedClients) == 0 and not
                self.game.world.gameIsInProgress()):
            # Don't shut down if local player is connected.
            for p in self.game.world.players:
                if not p.bot:
                    break
            else:
                # Shut down the server.
                self.shutdown()

    def shutdown(self):
        if self._alreadyShutdown:
            return
        self._alreadyShutdown = True

        # Kill server
        self.running = False
        self.game.stop()
        self.onShutdown.execute()



'''
leaderboard.py - defines the LeaderBoard class which deals with drawing the
leader board to the screen.
'''

import pygame
import logging

from trosnoth.data import getPath, sprites
import trosnoth.gui.framework.framework as framework
from trosnoth.gui.common import (
    Location, Scalar, FullScreenAttachedPoint, SizedImage, ScaledSize)
from trosnoth.gui.framework import table
from trosnoth.gui.framework.elements import PictureElement, SolidRect
from trosnoth.model import gameStates
from trosnoth.utils.twist import WeakCallLater

log = logging.getLogger(__name__)

# How often the leaderboard will update (in seconds)
UPDATE_DELAY = 1.0
SHORT_UPGRADE_NAMES = {
    'Minimap Disruption': 'minimap',
    'Phase Shift': 'phase',
}
# Upgrades that the enemy team have that you are allowed to see
PUBLIC_UPGRADES = ['Minimap Disruption', 'Turret', 'Shield', 'Grenade']

class LeaderBoard(framework.CompoundElement):

    def __init__(self, app, game, gameViewer, shadow=False):
        super(LeaderBoard, self).__init__(app)
        self.world = game.world
        self.shadow = shadow
        self.gameViewer = gameViewer
        self.app = app
        self.scale = self.app.screenManager.scaleFactor

        self.xPos = 4
        self.yPos = (self.gameViewer.miniMap.getRect().bottom +
                self.gameViewer.zoneBarHeight + 5)
        position = Location(FullScreenAttachedPoint((self.xPos, self.yPos),
                'topright'), 'topright')

        # Create the table and set all the appropriate style attributes
        self.playerTable = table.Table(app, position, columns=4, topPadding=4)

        self.playerTable.setBorderWidth(0)
        self.playerTable.setBorderColour((0, 0, 0))

        self.playerTable.style.backColour = None
        self.playerTable.style.foreColour = app.theme.colours.leaderboardNormal
        self.playerTable.style.font = app.screenManager.fonts.leaderboardFont
        self.playerTable.style.padding = (4, 1)
        self.playerTable.style.hasShadow = shadow
        self.playerTable.style.shadowColour = (0, 0, 0)

        self.playerTable.getColumn(0).style.textAlign = 'midright'
        self.playerTable.getColumn(1).style.textAlign = 'midright'
        self.playerTable.getColumn(2).style.textAlign = 'center'
        self.playerTable.getColumn(3).style.textAlign = 'midleft'

        self.rowHeight = Scalar(25)

        self.playerTable.getColumn(0).setWidth(30)    # Flags
        self.playerTable.getColumn(1).setWidth(100)   # Name
        self.playerTable.getColumn(2).setWidth(25)    # Star
        self.playerTable.getColumn(3).setWidth(70)    # Upgrade

        self._resetPlayers()

        self.elements = [self.playerTable]

        bgColour = app.theme.colours.leaderboardBackground
        if bgColour[3]:
            background = SolidRect(
                app, bgColour, bgColour[3], self.playerTable)
            self.elements.insert(0, background)
        self.update(True)

    def _resetPlayers(self):
        self.players = dict((k, []) for k in self.world.teamWithId)

    def update(self, loop=False):
        self._resetPlayers()

        if self.world.gameIsInProgress():
            yPos = self.yPos
        else:
            yPos = self.yPos + int(50 * self.scale)
        self.playerTable.pos = Location(FullScreenAttachedPoint(
                (self.xPos, yPos), 'topright'), 'topright')

        try:
            pi = self.gameViewer.interface.runningPlayerInterface
            self.friendlyTeam = pi.player.team
            self.spectator = False
        except (AttributeError, KeyError):
            # This will occur if the player is not yet on a team
            self.friendlyTeam = self.world.teamWithId['A']
            self.spectator = True

        for player in self.world.players:
            # getDetails() returns dict with pID, nick, team, dead, stars,
            # upgrade
            details = player.getDetails()
            self.players[details['team']].append(details)

        self._sort('A')
        self._sort('B')
        self._sort('\x00')
        self._updateTable()

        if loop:
            self.callDef = WeakCallLater(UPDATE_DELAY, self, 'update', True)

    def _sort(self, teamId):
        sortedList = sorted(
                [p for p in self.players[teamId] if not p['dead']],
                key=lambda p: p['stars'], reverse=True)
        sortedList.extend([p for p in self.players[teamId] if p['dead']])
        self.players[teamId] = sortedList


    def _updateTable(self):
        for x in range(self.playerTable.rowCount()):
            self.playerTable.delRow(0)

        index = -1
        self.contents = []

        teamOrder = [self.friendlyTeam] + [t for t in self.world.teams if t !=
                self.friendlyTeam]
        if self.friendlyTeam is not None:
            teamOrder.append(None)
        for team in teamOrder:
            if team is None:
                teamId = '\x00'
                # If we're in the lobby, only show real players
                visiblePlayers = [p for p in self.players[teamId]
                        if not p['bot']]
            else:
                teamId = team.id
                visiblePlayers = self.players[teamId]

            if len(visiblePlayers) == 0:
                continue

            index = self._addTeamHeader(team, teamId, visiblePlayers, index)

            # Add a row for each player
            for player in visiblePlayers:
                index = self._addPlayerRow(team, teamId, player, index)

            # Add the 'star total' row
            index = self._addTeamFooter(team, teamId, visiblePlayers, index)

    def _addRow(self, index, height):
        index += 1
        self.playerTable.addRow()
        row = self.playerTable.getRow(index)
        row.setHeight(height)
        return index

    def _embedTeamHeaderStar(self, index):
        starImage = SizedImage(getPath(sprites, 'smallstar.png'),
                ScaledSize(14, 13), (255,255,255))
        xPos = int(1 * self.scale)
        yPos = int(5 * self.scale)
        starImage = PictureElement(self.app, starImage,
                Location(table.CellAttachedPoint((xPos,yPos),
                self.playerTable[2][index], 'topleft')))
        self.playerTable[2][index].elements = [starImage]

    def _writeTeamHeaderPlayerCount(self, players, index):
        if len(players) == 1:
            noun = 'player'
        else:
            noun = 'players'
        self.playerTable[3][index].setText(str(len(players)) + ' ' + noun)

    def _setTeamHeaderStatus(self, index, text, colour):
        cell = self.playerTable[3][index]
        cell.setText(text)
        cell.style.foreColour = colour

    def _addTeamHeader(self, team, teamId, players, index):
        index = self._addRow(index, self.rowHeight)
        self.playerTable[1][index].setText( self.world.getTeamName(teamId))

        if self.world.gameIsInProgress():
            self._embedTeamHeaderStar(index)
            self._writeTeamHeaderPlayerCount(players, index)
        else:
            self.playerTable[2][index].elements = []

        teamColour = self.app.theme.colours.leaderboard(team)
        if self.world.state.displayWinner():
            if team is None:
                pass
            elif self.world.getWinningTeam() == team:
                self._setTeamHeaderStatus(index, 'has won!', teamColour)
            elif self.world.getWinningTeam() is not None:
                self._setTeamHeaderStatus(index, 'has lost.', teamColour)
            else:
                self._setTeamHeaderStatus(index, 'has drawn!', teamColour)

        self.playerTable.getRow(index).style.foreColour = teamColour
        self.contents.append((index, 'header', team))

        return index

    def _addUpgradeText(self, team, player, index):
        if player['upgrade'] is None:
            upgradeText = ''
        elif (team != self.friendlyTeam and player['upgrade'] not in
                PUBLIC_UPGRADES and team is not None):
            upgradeText = ''
        else:
            if player['upgrade'] in SHORT_UPGRADE_NAMES:
                upgradeText = '+ ' + SHORT_UPGRADE_NAMES[player['upgrade']]
            else:
                upgradeText = '+ ' + player['upgrade'].lower()

        cell = self.playerTable[3][index]
        cell.setText(upgradeText)
        cell.style.foreColour = self.app.theme.colours.leaderboardNormal

    def _setPlayerColour(self, team, player, index):
        if player['dead']:
            colour = self.app.theme.colours.leaderboardDead
        else:
            colour = self.app.theme.colours.leaderboard(team)
        self.playerTable[1][index].style.foreColour = colour
        self.playerTable[2][index].style.foreColour = colour

    def _addPlayerRow(self, team, teamId, player, index):
        index = self._addRow(index, self.rowHeight)

        flags = ''
        if self.world.state == gameStates.Lobby:
            if player['ready']:
                flags = 'R'
            elif player['tournament']:
                flags = 'T'

        self.playerTable[0][index].setText(flags)
        self.playerTable[1][index].setText(player['nick'])

        if not self.world.state.isStartingCountdown():
            if team is self.friendlyTeam or self.spectator:
                self.playerTable[2][index].setText(str(player['stars']))

            self._addUpgradeText(team, player, index)

        self._setPlayerColour(team, player, index)

        return index

    def _addTeamFooter(self, team, teamId, players, index):
        if (len(players) > 1
                and not self.world.state.isStartingCountdown()
                and team is not None):
            index = self._addRow(index, self.rowHeight)
            self.playerTable[1][index].setText('Total:')
            teamStars = self.world.getTeamStars(team)
            self.playerTable[2][index].setText(teamStars)
            if teamStars == 1:
                noun = 'star'
            else:
                noun = 'stars'
            self.playerTable[3][index].setText(noun)
            self.contents.append((index, 'total', team))

        return index

    def kill(self):
        self.callDef.cancel()

    def draw(self, surface):
        super(LeaderBoard, self).draw(surface)

        # pointA = left-most point of the line
        # pointB = right-most point of the line
        # pointC = left-most point of the line shadow
        # pointD = right-most point of the line shadow

        for (index, rowKind, team) in self.contents:
            if rowKind == 'header':
                rightMargin = int(8 * self.scale)
                leftMargin = int(10 * self.scale)
                bottomMargin = int(2 * self.scale)

                point = self.playerTable._getRowPt(index + 1)
                pointA = (point[0] + leftMargin, point[1] - bottomMargin)
                pointB = (pointA[0] + self.playerTable._getSize()[0] -
                        leftMargin - rightMargin, pointA[1])
                teamColour = self.app.theme.colours.leaderboard(team)
                pygame.draw.line(surface, teamColour, pointA, pointB, 1)
                if self.shadow:
                    pointC, pointD = shift(pointA, pointB)
                    pygame.draw.line(surface, (0, 0, 0), pointC, pointD, 1)

                pointA, pointB = shift(pointA, pointB,
                        -self.playerTable.getRow(index)._getHeight() +
                        bottomMargin, 0)
                teamColour = self.app.theme.colours.leaderboard(team)
                pygame.draw.line(surface, teamColour, pointA, pointB, 1)
                if self.shadow:
                    pointC, pointD = shift(pointA, pointB)
                    pygame.draw.line(surface, (0, 0, 0), pointC, pointD, 1)

            elif rowKind == 'total':
                bottomMargin = int(2 * self.scale)

                rowPoint = self.playerTable._getRowPt(index)
                colPoint = self.playerTable._getColPt(2)
                pointA = (colPoint[0], rowPoint[1] - bottomMargin)
                pointB = (pointA[0] + self.playerTable.getColumn(2)._getWidth(),
                        pointA[1])
                pygame.draw.line(surface,
                        self.app.theme.colours.leaderboardNormal, pointA,
                        pointB, 1)
                if self.shadow:
                    pointC, pointD = shift(pointA, pointB)
                    pygame.draw.line(surface, (0, 0, 0), pointC, pointD, 1)

def shift(pointA, pointB, down = 1, right = 1):
    '''Returns two points that are shifted by the specified number of pixels.
    Negative values shift up and left, positive values shift down and right.'''
    return ((pointA[0] + right, pointA[1] + down),
            (pointB[0] + right, pointB[1] + down))




import mox
import logging
import StringIO
import sys

def pytest_funcarg__mock(request):
    mock = mox.Mox()
    request.addfinalizer(lambda : mock.UnsetStubs())
    return mock

def sendMessage(plug, msg):
    plug._receive(msg)

def pytest_funcarg__logman(request):
    '''
    Defines a log manager py.test funcarg which redirects logs to stdout and
    allows assertions about whether errors were recorded in the logs.
    '''
    logman = LogManager()
    request.addfinalizer(logman.teardown)
    return logman

class LogManager(object):
    def __init__(self):
        self.stdout_log_handler = logging.StreamHandler(sys.stdout)
        self.stdout_log_handler.setFormatter(logging.Formatter('%(levelname)s:%(name)s: %(message)s'))
        self.error_buffer = StringIO.StringIO()
        self.error_log_handler = logging.StreamHandler(self.error_buffer)
        self.expect_no_errors = True

        self.stdout_log_handler.setLevel(logging.DEBUG)
        logging.getLogger().addHandler(self.stdout_log_handler)
        self.error_log_handler.setLevel(logging.ERROR)
        logging.getLogger().addHandler(self.error_log_handler)

        logging.getLogger().setLevel(logging.DEBUG)

    def ignore_errors(self):
        '''
        Call this method to instruct the log manager to ignore ERROR or CRITICAL
        log messages. Note that if you have already called assert_errors() or
        assert_no_errors(), you do not need to call ignore_errors().
        '''
        self.expect_no_errors = False

    def teardown(self):
        logging.getLogger().removeHandler(self.stdout_log_handler)
        logging.getLogger().removeHandler(self.error_log_handler)
        if self.expect_no_errors:
            self.assert_no_errors()

    def assert_no_errors(self):
        assert self.error_buffer.len == 0, 'There were errors in the log'
        self.expect_no_errors = False

    def assert_errors(self):
        assert self.error_buffer.len > 0, 'No errors in logs'
        self.expect_no_errors = False



from collections import defaultdict
import logging

try:
    import simplejson as json
except ImportError:
    import json

from twisted.internet import reactor

from trosnoth.const import (
    SHOT_DEATH, TURRET_DEATH, GRENADE_DEATH, SHOXWAVE_DEATH,
)
from trosnoth.model.upgrades import allUpgrades
from trosnoth.model.zone import ZoneState
from trosnoth.utils.event import Event

log = logging.getLogger(__name__)


##################################
# Collection of all achievements #
##################################

class AchievementSet(object):
    def __init__(self, template=None):
        self.all = []
        self.byId = {}
        self.oncePerGame = []
        self.oncePerTeamPerGame = []
        self.perPlayer = []
        if template:
            for achievement in template.all:
                self.register(achievement)

    def getAchievementDetails(self, idstring):
        achievement = self.byId[idstring]
        if achievement.name == "":
            return idstring, achievement.description
        else:
            return achievement.name, achievement.description

    def register(self, achievement):
        if achievement.idstring in self.byId:
            raise KeyError('achievement with id %r already registered' % (
                    achievement.idstring,))

        self.all.append(achievement)
        self.byId[achievement.idstring] = achievement

        if achievement.oncePerGame:
            self.oncePerGame.append(achievement)
        elif achievement.oncePerTeamPerGame:
            self.oncePerTeamPerGame.append(achievement)
        else:
            assert achievement.perPlayer
            self.perPlayer.append(achievement)
        return achievement

availableAchievements = AchievementSet()


############################
# Achievement base classes #
############################

class Achievement(object):

    keepProgress = False
    afterGame = False
    inGame = True

    oncePerGame = False
    oncePerTeamPerGame = False

    def __init__(self, world):
        self.world = world
        self.unlocked = False
        self.registered = False
        self.onUnlocked = Event()

    def save(self):
        return False

    def start(self):
        if self.isActive():
            self.register()

    def stop(self):
        if self.registered:
            self.unregister()

    def register(self):
        self.registered = True

    def unregister(self):
        self.registered = False

    def _sendUnlocked(self, player):
        self.onUnlocked(self, player)
        log.debug(self.achievementString(player))

    def achievementString(self, player):
        return 'Achievement unlocked by %s! - %s' % (
            player.nick, self.name if self.name else self.idstring)

    def __str__(self):
        return '%s' % self.idstring

    @classmethod
    def describe(cls):
        '''
        Used when saving achievement meta-information to file. Should be
        overwritten by higher level classes.
        '''
        information = {'name': cls.name,
                       'description': cls.description,
                       'type': 'boolean'}
        return {cls.idstring: information}

    def isActive(self):
        return not self.unlocked


class PlayerAchievement(Achievement):
    perPlayer = True

    def __init__(self, player, world, *args, **kwargs):
        super(PlayerAchievement, self).__init__(world, *args, **kwargs)
        self.player = player
        self.world = world

    def rejoined(self, player):
        if self.registered:
            self.unregister()
            self.player = player
            self.register()
        else:
            self.player = player

    def achievementTriggered(self):
        gameOn = self.world.gameIsInProgress()
        if (gameOn and not self.inGame) or (not gameOn and not self.afterGame):
            return

        self.unlocked = True
        self._sendUnlocked(self.player)
        if self.registered and not self.isActive():
            self.unregister()


class PlayerlessAchievement(Achievement):
    perPlayer = False

    def achievementTriggeredFor(self, players):
        if self.unlocked:
            return
        gameOn = self.world.gameIsInProgress()
        if (gameOn and not self.inGame) or (not gameOn and not self.afterGame):
            return

        self.unlocked = True
        for player in players:
            self._sendUnlocked(player)
        if self.registered and not self.isActive():
            self.unregister()

    def isActive(self):
        # By default, there's no need to stay subscribed to events once the
        # achievement is achieved.
        return not self.unlocked


class OncePerGame(PlayerlessAchievement):
    oncePerGame = True


class OncePerTeamPerGame(PlayerlessAchievement):
    oncePerTeamPerGame = True

    def __init__(self, world, team, *args, **kwargs):
        super(OncePerTeamPerGame, self).__init__(world, *args, **kwargs)
        self.team = team


class OncePerPlayerPerGame(PlayerAchievement):
    def achievementTriggered(self):
        if not self.unlocked:
            super(OncePerPlayerPerGame, self).achievementTriggered()


class NoLimit(PlayerAchievement):
    def isActive(self):
        return True


class OnceEverPerPlayer(PlayerAchievement):
    '''
    self.unlocked is "unlocked this time".
    self.previouslyUnlocked is "unlocked previously".
    '''
    keepProgress = True

    def __init__(self, *args, **kwargs):
        super(OnceEverPerPlayer, self).__init__(*args, **kwargs)
        self.previouslyUnlocked = False
        self.dbObject = None
        self.previouslyUnlocked = False
        self.progress = 0
        self.data = None

        if self.player is not None and not self.player.bot:
            # Delay this call so that the auth user can be set
            reactor.callLater(0, self.readExistingData)

    def save(self):
        if self.dbObject is None:
            return False
        if self.progress != self.dbObject.progress:
            self.dbObject.progress = self.progress
            self.dbObject.unlocked = self.unlocked or self.previouslyUnlocked
            self.dbObject.save()
            return True
        return False

    def readExistingData(self):
        user = self.player.user
        if user is None:
            return

        self.dbObject = record = user.getAchievementRecord(self.idstring)
        if record.data:
            self.data = json.loads(record.data)
        else:
            self.data = None
        self.previouslyUnlocked = record.unlocked
        self.progress = record.progress

    def achievementTriggered(self):
        if not self.unlocked and not self.previouslyUnlocked:
            super(OnceEverPerPlayer, self).achievementTriggered()

    def isActive(self):
        return not self.unlocked and not self.previouslyUnlocked

    def getJsonString(self):
        return {
            'unlocked' : self.unlocked or self.previouslyUnlocked,
            'progress' : self.progress,
        }


###################
# Streak subclass #
###################

class Streak(PlayerAchievement):
    streakTarget = -1

    def __init__(self, *args, **kwargs):
        self.progress = 0
        super(Streak, self).__init__(*args, **kwargs)

    def increment(self, amount=1):
        gameOn = self.world.gameIsInProgress()
        if (gameOn and not self.inGame) or (not gameOn and not self.afterGame):
            return
        self.progress = min(self.streakTarget, self.progress + amount)

        if self.progress == self.streakTarget:
            self.achievementTriggered()

    def reset(self):
        self.progress = 0

    def __str__(self):
        return "%s: %d/%d" % (
            self.idstring, self.progress, self.streakTarget)

    @classmethod
    def describe(self):
        information = {'name': self.name,
                       'description': self.description,
                       'type': 'incremental',
                       'requirements': self.streakTarget,
                       'keepProgress': self.keepProgress}
        return {self.idstring: information}


##########################
# Achievement subclasses #
##########################

class ChecklistAchievement(Achievement):
    requiredItems = set()

    def __init__(self, *args, **kwargs):
        super(ChecklistAchievement, self).__init__(*args, **kwargs)

    def addItem(self, item):
        gameOn = self.world.gameIsInProgress()
        if (gameOn and not self.inGame) or (not gameOn and not self.afterGame):
            return
        if not self.data:
            self.data = set()
        self.data.add(item)

        if self.requiredItems == self.data:
            self.achievementTriggered()

    def __str__(self):
        return "%s: %d/%d items" % (self.idstring, len(self.data),
                len(self.requiredItems))

    @classmethod
    def describe(self):
        information = {'name': self.name,
                       'description': self.description,
                       'type': 'checklist',
                       'requirements': list(self.requiredItems),
                       'keepProgress': self.keepProgress}
        return {self.idstring: information}


class PersistedChecklistAchievement(ChecklistAchievement, OnceEverPerPlayer):
    def save(self):
        if self.dbObject is None:
            return False
        changed = False
        if self.progress != self.dbObject.progress:
            self.dbObject.progress = self.progress
            changed = True

        data = json.dumps(list(self.data))
        if data != self.dbObject.data:
            self.dbObject.data = data
            changed = True

        if changed:
            self.dbObject.unlocked = self.unlocked or self.previouslyUnlocked
            self.dbObject.save()

        return changed

    def readExistingData(self):
        super(PersistedChecklistAchievement, self).readExistingData()
        self.data = set(self.data or [])


class TimedAchievement(PlayerAchievement):
    requiredValue = -1
    timeWindow = -1

    def __init__(self, *args, **kwargs):
        super(TimedAchievement, self).__init__(*args, **kwargs)
        self.rollingList = []

    def addToList(self):
        now = self.world.timer.getElapsedTime()
        self.rollingList.append(now)

        while (self.rollingList and (
                now - self.rollingList[0]) > self.timeWindow):
            del self.rollingList[0]

        if len(self.rollingList) >= self.requiredValue:
            self.achievementTriggered()
            self.rollingList = []


#########################
# Concrete Achievements #
#########################

@availableAchievements.register
class Neutraliser(NoLimit):
    idstring = 'neutraliser'
    name = 'Neutraliser'
    description = 'Neutralise three zones with a single tag'

    def register(self):
        super(Neutraliser, self).register()
        self.player.onNeutralisedSector.addListener(self.neutralisedSector)

    def unregister(self):
        super(Neutraliser, self).unregister()
        self.player.onNeutralisedSector.removeListener(self.neutralisedSector)

    def neutralisedSector(self, zoneCount):
        if zoneCount >= 3:
            self.achievementTriggered()


class MultiTagBase(Streak, OncePerPlayerPerGame):
    def register(self):
        super(MultiTagBase, self).register()
        self.player.onTaggedZone.addListener(self.taggedZone)
        self.player.onDied.addListener(self.died)

    def unregister(self):
        super(MultiTagBase, self).unregister()
        self.player.onTaggedZone.removeListener(self.taggedZone)
        self.player.onDied.removeListener(self.died)

    def taggedZone(self, *args, **kwargs):
        self.increment()

    def died(self, *args, **kwargs):
        self.reset()


@availableAchievements.register
class MultiTagSmall(MultiTagBase):
    idstring = 'multiTagSmall'
    name = 'I Ain\'t Even Winded'
    description = 'Capture three zones in a single life'
    streakTarget = 3


@availableAchievements.register
class MultiTagMedium(MultiTagBase):
    idstring = 'multiTagMedium'
    name = 'The Long Way Home'
    description = 'Capture five zones in a single life'
    streakTarget = 5


@availableAchievements.register
class MultiTagLarge(MultiTagBase):
    idstring = 'multiTagLarge'
    name = 'Cross-Country Marathon'
    description = 'Capture eight zones in a single life'
    streakTarget = 8


@availableAchievements.register
class Janitor(Streak, OncePerPlayerPerGame):
    idstring = 'neutralCapturesMedium'
    name = 'Janitor'
    description = (
        'Capture 5 neutral zones away from enemy territory in a single game')
    streakTarget = 5

    def register(self):
        super(Janitor, self).register()
        self.player.onTaggedZone.addListener(self.taggedZone)

    def unregister(self):
        super(Janitor, self).unregister()
        self.player.onTaggedZone.removeListener(self.taggedZone)

    def taggedZone(self, zone, previousOwner):
        if previousOwner is None and not any(
                self.player.isEnemyTeam(z.owner)
                for z in zone.getAdjacentZones()):
            self.increment()


class TotalTagsBase(Streak, OnceEverPerPlayer):
    def register(self):
        super(TotalTagsBase, self).register()
        self.player.onTaggedZone.addListener(self.taggedZone)

    def unregister(self):
        super(TotalTagsBase, self).unregister()
        self.player.onTaggedZone.removeListener(self.taggedZone)

    def taggedZone(self, *args, **kwargs):
        self.increment()


@availableAchievements.register
class TotalTagsSmall(TotalTagsBase):
    idstring = 'totalTagsSmall'
    name = 'Cultural Assimilation'
    description = 'Capture 20 zones'
    streakTarget = 20


@availableAchievements.register
class TotalTagsMedium(TotalTagsBase):
    idstring = 'totalTagsMedium'
    name = 'Globalization'
    description = 'Capture 50 zones'
    streakTarget = 50


class MultiKillBase(Streak, OncePerPlayerPerGame):
    def register(self):
        super(MultiKillBase, self).register()
        self.player.onKilled.addListener(self.killedPlayer)
        self.player.onDied.addListener(self.died)

    def unregister(self):
        super(MultiKillBase, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)
        self.player.onDied.removeListener(self.died)

    def killedPlayer(self, *args, **kwargs):
        self.increment()

    def died(self, *args, **kwargs):
        self.reset()


@availableAchievements.register
class MultiKillSmall(MultiKillBase):
    idstring = 'multikillSmall'
    name = 'Triple Threat'
    description = 'Kill three enemies in a single life'
    streakTarget = 3


@availableAchievements.register
class MultiKillMedium(MultiKillBase):
    idstring = 'multikillMedium'
    name = 'High Five'
    description = 'Kill five enemies in a single life'
    streakTarget = 5


@availableAchievements.register
class MultiKillLarge(MultiKillBase):
    idstring = 'multikillLarge'
    name = "That's the Badger"
    description = 'Kill nine enemies in a single life'
    streakTarget = 9


class TotalKillsBase(Streak, OnceEverPerPlayer):
    def register(self):
        super(TotalKillsBase, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(TotalKillsBase, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, *args, **kwargs):
        self.increment()


@availableAchievements.register
class TotalKillsSmall(TotalKillsBase):
    idstring = 'totalKillsSmall'
    name = 'Family Friendly Fun'
    description = 'Kill 50 enemies'
    streakTarget = 50


@availableAchievements.register
class TotalKillsMedium(TotalKillsBase):
    idstring = 'totalKillsMedium'
    name = "All in a Day's Work"
    description = 'Kill 100 enemies'
    streakTarget = 100


@availableAchievements.register
class ShoppingSpree(Streak, OncePerPlayerPerGame):
    idstring = 'multiUpgradesSmall'
    name =  'Shopping Spree'
    description = 'Buy two upgrades in a single life'
    streakTarget = 2

    def register(self):
        super(ShoppingSpree, self).register()
        self.player.onUsedUpgrade.addListener(self.usedUpgrade)
        self.player.onDied.addListener(self.died)

    def unregister(self):
        super(ShoppingSpree, self).unregister()
        self.player.onUsedUpgrade.removeListener(self.usedUpgrade)
        self.player.onDied.removeListener(self.died)

    def usedUpgrade(self, *args, **kwargs):
        self.increment()

    def died(self, *args, **kwargs):
        self.reset()


@availableAchievements.register
class SmartShopper(Streak, OncePerPlayerPerGame):
    idstring = 'multiUpgradesMedium'
    name = 'Smart Shopper'
    description = 'Buy five upgrades in a single game'
    streakTarget = 5

    def register(self):
        super(SmartShopper, self).register()
        self.player.onUsedUpgrade.addListener(self.usedUpgrade)

    def unregister(self):
        super(SmartShopper, self).unregister()
        self.player.onUsedUpgrade.removeListener(self.usedUpgrade)

    def usedUpgrade(self, *args, **kwargs):
        self.increment()


@availableAchievements.register
class BulletsSmall(Streak, OncePerPlayerPerGame):
    idstring = 'bulletsSmall'
    name = 'Machine Gunner'
    description = 'Shoot 100 bullets in a single life'
    streakTarget = 100

    def register(self):
        super(BulletsSmall, self).register()
        self.player.onShotFired.addListener(self.shotFired)
        self.player.onDied.addListener(self.died)

    def unregister(self):
        super(BulletsSmall, self).unregister()
        self.player.onShotFired.removeListener(self.shotFired)
        self.player.onDied.removeListener(self.died)

    def shotFired(self, *args, **kwargs):
        self.increment()

    def died(self, *args, **kwargs):
        self.reset()


@availableAchievements.register
class BulletsMedium(Streak, OncePerPlayerPerGame):
    idstring = 'bulletsMedium'
    name = 'Ammunition Overdrive'
    description = 'Shoot 500 bullets in a single game'
    streakTarget = 500

    def register(self):
        super(BulletsMedium, self).register()
        self.player.onShotFired.addListener(self.shotFired)

    def unregister(self):
        super(BulletsMedium, self).unregister()
        self.player.onShotFired.removeListener(self.shotFired)

    def shotFired(self, *args, **kwargs):
        self.increment()


class WinnerBase(Streak, OnceEverPerPlayer):
    def register(self):
        super(WinnerBase, self).register()
        self.world.onGameOver.addListener(self.gameOver)

    def unregister(self):
        super(WinnerBase, self).unregister()
        self.world.onGameOver.removeListener(self.gameOver)

    def gameOver(self, team, *args, **kwarsg):
        if not self.player.isFriendsWithTeam(team):
            return

        playerCount = sum(1 for p in self.world.players if not p.bot)
        if playerCount >= 6:
            self.increment()


@availableAchievements.register
class WinnerTiny(WinnerBase):
    idstring = 'winnerTiny'
    name = 'Trosnoth Newbie'
    description = 'Win your first game of Trosnoth (minimum 6 players)'
    streakTarget = 1


@availableAchievements.register
class WinnerSmall(WinnerBase):
    idstring = 'winnerSmall'
    name = 'Trosnoth Amateur'
    description = 'Win 5 games of Trosnoth (minimum 6 players)'
    streakTarget = 5


@availableAchievements.register
class WinnerMedium(WinnerBase):
    idstring = 'winnerMedium'
    name = 'Trosnoth Consultant'
    description = 'Win 10 games of Trosnoth (minimum 6 players)'
    streakTarget = 10


@availableAchievements.register
class WinnerLarge(WinnerBase):
    idstring = 'winnerLarge'
    name = 'Trosnoth Professional'
    description = 'Win 20 games of Trosnoth (minimum 6 players)'
    streakTarget = 20


@availableAchievements.register
class WinnerHuge(WinnerBase):
    idstring = 'winnerHuge'
    name = 'Trosnoth Expert'
    description = 'Win 50 games of Trosnoth (minimum 6 players)'
    streakTarget = 50


@availableAchievements.register
class AssistTags(Streak, OncePerPlayerPerGame):
    idstring = 'assistTags'
    name = 'Credit to Team'
    description = 'Assist in the tagging of five zones in a single game'
    streakTarget = 5

    def register(self):
        super(AssistTags, self).register()
        self.world.onZoneTagged.addListener(self.zoneTagged)

    def unregister(self):
        super(AssistTags, self).unregister()
        self.world.onZoneTagged.removeListener(self.zoneTagged)

    def zoneTagged(self, zone, player, *args, **kwargs):
        if player and self.player.isFriendsWith(player):
            if player == self.player:
                return
            if self.player in zone.players and not self.player.dead:
                self.increment()


class QuickKillBase(TimedAchievement):
    deathTypes = ''

    def register(self):
        super(QuickKillBase, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(QuickKillBase, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, target, deathType, *args, **kwargs):
        if deathType in self.deathTypes:
            self.addToList()


@availableAchievements.register
class QuickKillSmall(QuickKillBase, NoLimit):
    requiredValue = 2
    timeWindow = 5
    deathTypes = {SHOT_DEATH, TURRET_DEATH, SHOXWAVE_DEATH}
    idstring = 'quickKillSmall'
    name = 'Double Kill'
    description = 'Kill two enemies within five seconds (no grenades)'


@availableAchievements.register
class QuickKillMedium(QuickKillBase, NoLimit):
    requiredValue = 3
    timeWindow = 5
    deathTypes = {SHOT_DEATH, TURRET_DEATH, SHOXWAVE_DEATH}
    idstring = 'quickKillMedium'
    name = 'Triple Kill'
    description = 'Kill three enemies within five seconds (no grenades)'


@availableAchievements.register
class GrenadeMultiKill(QuickKillBase, NoLimit):
    requiredValue = 3
    # Processing of achievements should happen within 0.25 secs
    timeWindow = 0.25
    deathTypes = {GRENADE_DEATH}
    idstring = 'grenadeMultikill'
    name = "It's Super Effective!"
    description = 'Kill three enemies with a single grenade'


@availableAchievements.register
class RicochetAchievement(OncePerPlayerPerGame):
    name = 'Bouncy Flag'
    description = (
        'Kill an enemy with a bullet that has ricocheted at least once')
    idstring = 'ricochetKill'

    def register(self):
        super(RicochetAchievement, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(RicochetAchievement, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, target, deathType, *args, **kwargs):
        if deathType == SHOT_DEATH:
            if self.player.hasRicochet:
                self.achievementTriggered()


@availableAchievements.register
class AccuracySmall(OncePerPlayerPerGame):
    name = 'Boom, Headshot'
    description = 'Have an accuracy of 10% or higher during a game'
    idstring = 'accuracySmall'


@availableAchievements.register
class AccuracyMedium(OncePerPlayerPerGame):
    name = "Sniping's a Good Job, Mate"
    description = 'Have an accuracy of 15% or higher during a game'
    idstring = 'accuracyMedium'


@availableAchievements.register
class AccuracyLarge(OncePerPlayerPerGame):
    name = 'Professionals Have Standards'
    description = 'Have an accuracy of 20% or higher during a game'
    idstring = 'accuracyLarge'


@availableAchievements.register
class StatScore(OncePerPlayerPerGame):
    name = 'Master of the Internet'
    description = 'Reach a stats score of 1337 or higher during a game'
    idstring = 'statScore'


@availableAchievements.register
class TotalAliveTime(OnceEverPerPlayer):
    name = 'Long Live the King'
    description = 'Rack up a total of 1000 seconds of alive time'
    idstring = 'totalAliveTime'


@availableAchievements.register
class StayingAlive(OncePerPlayerPerGame):
    name = 'Never Asleep on the Job'
    description = 'Be alive at least 75% of the time'
    idstring = 'stayingAlive'


@availableAchievements.register
class AliveStreak(OncePerPlayerPerGame):
    name = 'Still Alive'
    description = 'Stay alive for at least 180 seconds in a single life'
    idstring = 'aliveStreak'


@availableAchievements.register
class UseStarsEfficiently(OncePerPlayerPerGame):
    name = 'Waste Not, Want Not'
    description = ('Use or contribute at least 50% of the stars you earn '
            'during a game')
    idstring = 'useStarsEfficiently'


@availableAchievements.register
class LongRangeKill(OncePerPlayerPerGame):
    name = 'Long-Range Ballistics'
    description = 'Hit an enemy at the maximum range of your gun'
    idstring = 'longRangeKill'

    def register(self):
        super(LongRangeKill, self).register()
        self.player.onShotHurtPlayer.addListener(self.hurtPlayer)

    def unregister(self):
        super(LongRangeKill, self).unregister()
        self.player.onShotHurtPlayer.removeListener(self.hurtPlayer)

    def hurtPlayer(self, target, shot):
        if shot.timeLeft < 0.1:
            self.achievementTriggered()


@availableAchievements.register
class TagsAndKills(OncePerPlayerPerGame):
    name = 'All-Rounder'
    description = 'Kill 10 people and tag 5 zones in a single life'
    idstring = 'tagsAndKills'

    requiredTags = 5
    requiredKills = 10

    def __init__(self, *args, **kwargs):
        super(TagsAndKills, self).__init__(*args, **kwargs)
        self.tags = 0
        self.kills = 0

    def register(self):
        super(TagsAndKills, self).register()
        self.player.onTaggedZone.addListener(self.taggedZone)
        self.player.onKilled.addListener(self.killedPlayer)
        self.player.onDied.addListener(self.died)

    def unregister(self):
        super(TagsAndKills, self).unregister()
        self.player.onTaggedZone.removeListener(self.taggedZone)
        self.player.onKilled.removeListener(self.killedPlayer)
        self.player.onDied.removeListener(self.died)

    def taggedZone(self, *args, **kwargs):
        self.tags = min(self.tags + 1, self.requiredTags)
        self.checkTrigger()

    def killedPlayer(self, *args, **kwargs):
        self.kills = min(self.kills + 1, self.requiredKills)
        self.checkTrigger()

    def checkTrigger(self):
        if self.kills == self.requiredKills and self.tags == self.requiredTags:
            self.achievementTriggered()

    def died(self, *args, **kwargs):
        self.kills = 0
        self.tags = 0


@availableAchievements.register
class KillEnemyWithStars(OncePerPlayerPerGame):
    name = 'Stop Right There, Criminal Scum'
    description = 'Kill an enemy holding five stars or more'
    idstring = 'killEnemyWithStars'

    def register(self):
        super(KillEnemyWithStars, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(KillEnemyWithStars, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, target, *args, **kwargs):
        if target.stars >= 5:
            self.achievementTriggered()


@availableAchievements.register
class DestroyStars(Streak, OnceEverPerPlayer):
    idstring = 'destroyStars'
    name = 'The Recession we Had to Have'
    description = 'Destroy 100 stars by killing the players carrying them'
    streakTarget = 100

    def register(self):
        super(DestroyStars, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(DestroyStars, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, target, *args, **kwargs):
        self.increment(target.stars)


@availableAchievements.register
class EarnStars(Streak, OnceEverPerPlayer):
    idstring = 'earnStars'
    name = 'Stimulus Package'
    description = 'Earn 100 stars'
    streakTarget = 100

    def __init__(self, *args, **kwargs):
        super(EarnStars, self).__init__(*args, **kwargs)
        self.lastStars = self.player.stars

    def register(self):
        super(EarnStars, self).register()
        self.player.onStarsChanged.addListener(self.starsChanged)

    def unregister(self):
        super(EarnStars, self).unregister()
        self.player.onStarsChanged.removeListener(self.starsChanged)

    def starsChanged(self, *args, **kwargs):
        if self.player.stars > self.lastStars:
            self.increment(self.player.stars - self.lastStars)
        self.lastStars = self.player.stars


@availableAchievements.register
class BuyEveryUpgrade(PersistedChecklistAchievement):
    idstring = 'buyEveryUpgrade'
    name = 'Technology Whiz'
    description = 'Use every upgrade at least once'
    requiredItems = set(u.upgradeType for u in allUpgrades)

    def register(self):
        super(BuyEveryUpgrade, self).register()
        self.player.onUsedUpgrade.addListener(self.usedUpgrade)

    def unregister(self):
        super(BuyEveryUpgrade, self).unregister()
        self.player.onUsedUpgrade.removeListener(self.usedUpgrade)

    def usedUpgrade(self, upgrade, *args, **kwargs):
        self.addItem(upgrade.upgradeType)


@availableAchievements.register
class UseMinimapDisruption(OncePerPlayerPerGame):
    idstring = 'minimapDisruption'
    name = 'The Rader Appears to be... Jammed!'
    description = 'Use a minimap disruption'

    def register(self):
        super(UseMinimapDisruption, self).register()
        self.player.onUsedUpgrade.addListener(self.usedUpgrade)

    def unregister(self):
        super(UseMinimapDisruption, self).unregister()
        self.player.onUsedUpgrade.removeListener(self.usedUpgrade)

    def usedUpgrade(self, upgrade, *args, **kwargs):
        if upgrade.upgradeType == 'm':
            self.achievementTriggered()


#disabled @availableAchievements.register
class AbandonEarly(OncePerPlayerPerGame):
    idstring = 'abandonEarly'
    name = 'Reduce, Reuse, Recycle'
    description = 'Abandon a turret or phase shift before it runs out'

    def register(self):
        super(AbandonEarly, self).register()
        self.player.onAbandonUpgrade.addListener(self.abandonUpgrade)

    def unregister(self):
        super(AbandonEarly, self).unregister()
        self.player.onAbandonUpgrade.removeListener(self.abandonUpgrade)

    def abandonUpgrade(self, upgrade, *args, **kwargs):
        '''
        Don't award if it was within 3 seconds of running out anyway.
        '''
        if upgrade.upgradeType in 'th' and self.player.upgradeGauge >= 3:
            self.achievementTriggered()


@availableAchievements.register
class GoodManners(OncePerPlayerPerGame):
    idstring = 'goodManners'
    name = 'Good Manners'
    description = 'Say "gg" after the game finishes'
    afterGame = True
    inGame = False

    def register(self):
        super(GoodManners, self).register()
        self.world.onOpenChatReceived.addListener(self.chat)

    def unregister(self):
        super(GoodManners, self).unregister()
        self.world.onOpenChatReceived.removeListener(self.chat)

    def chat(self, text, sender, *args, **kwargs):
        if text.lower().strip() == 'gg' and sender == self.player:
            self.achievementTriggered()


@availableAchievements.register
class DarkZoneKills(Streak, OncePerPlayerPerGame):
    idstring = 'darkZoneKills'
    name = 'Behind Enemy Lines'
    description = "Kill five enemies when you're in a dark zone of their colour"
    streakTarget = 5

    def register(self):
        super(DarkZoneKills, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(DarkZoneKills, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, *args, **kwargs):
        if self.player.team is None:
            return
        zone = self.player.getZone()
        if not zone:
            return
        if (self.player.isEnemyTeam(zone.owner) and zone.dark
                and not self.player.dead):
            self.increment()


class UpgradeKillBase(OncePerPlayerPerGame):
    def register(self):
        super(UpgradeKillBase, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(UpgradeKillBase, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, target, *args, **kwargs):
        upgrade = target.upgrade
        if upgrade and upgrade.upgradeType in self.upgradeTypes:
            self.achievementTriggered()


@availableAchievements.register
class NinjaKill(UpgradeKillBase):
    idstring = 'ninjaKill'
    name = "Ninjas Can't Catch You if You're on Fire"
    description = 'Kill an invisible ninja'
    upgradeTypes = 'n'


@availableAchievements.register
class AntiDisruption(UpgradeKillBase):
    idstring = 'antiDisruption'
    name = "Back Online"
    description = 'Kill an enemy who is using a Minimap Disruption'
    upgradeTypes = 'm'


@availableAchievements.register
class MinimapDisruptionTag(OncePerPlayerPerGame):
    name = 'Under the Radar'
    description = 'Tag a zone while the enemy\'s minimap is disrupted'
    idstring = 'disruptionTag'

    def register(self):
        super(MinimapDisruptionTag, self).register()
        self.player.onTaggedZone.addListener(self.taggedZone)

    def unregister(self):
        super(MinimapDisruptionTag, self).unregister()
        self.player.onTaggedZone.removeListener(self.taggedZone)

    def taggedZone(self, *args, **kwargs):
        for player in self.world.players:
            if (
                    self.player.isFriendsWith(player) and player.upgrade
                    and player.upgrade.upgradeType == 'm'):
                self.achievementTriggered()
                break


@availableAchievements.register
class ChokepointSpam(Streak, OncePerPlayerPerGame):
    idstring = 'chokepointSpam'
    name = "Chokepoint Checkmate"
    description = 'Kill three players without moving or dying (no upgrades)'
    streakTarget = 3

    pos = (0, 0)

    def register(self):
        super(ChokepointSpam, self).register()
        self.player.onKilled.addListener(self.killedPlayer)
        self.player.onShotFired.addListener(self.shotFired)

    def unregister(self):
        super(ChokepointSpam, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)
        self.player.onShotFired.removeListener(self.shotFired)

    def shotFired(self, *args, **kwargs):
        # Strictly speaking, the player is still able to move or die, but if
        # they don't shoot from exactly the same spot it won't count anyway.
        if self.player.pos != self.pos:
            self.reset()
        self.pos = self.player.pos

    def killedPlayer(self, shooter, deathType, *args, **kwargs):
        if self.player.upgrade is None and deathType != GRENADE_DEATH:
            self.increment()


#disabled @availableAchievements.register
class PhaseShiftAbsorb(Streak, OncePerPlayerPerGame):
    idstring = 'phaseShiftAbsorb'
    name = "Can't Touch This"
    description = 'Get hit by 30 bullets in a single phase shift'
    streakTarget = 30

    def register(self):
        super(PhaseShiftAbsorb, self).register()
        self.player.onHitByShot.addListener(self.hitByShot)
        self.player.onUpgradeGone.addListener(self.upgradeGone)

    def unregister(self):
        super(PhaseShiftAbsorb, self).unregister()
        self.player.onHitByShot.removeListener(self.hitByShot)
        self.player.onUpgradeGone.removeListener(self.upgradeGone)

    def hitByShot(self, *args, **kwargs):
        if self.player.upgrade and self.player.upgrade.upgradeType == 'h':
            self.increment()

    def upgradeGone(self, upgrade, *args, **kwargs):
        if upgrade.upgradeType == 'h':
            self.reset()


@availableAchievements.register
class KillAsRabbit(OncePerPlayerPerGame):
    name = 'Never Surrender'
    description = 'Kill an enemy after losing a game'
    idstring = 'killAsRabbit'
    afterGame = True
    inGame = False

    def register(self):
        super(KillAsRabbit, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(KillAsRabbit, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, *args, **kwargs):
        if (self.player.team is not None and
                self.player.isEnemyTeam(self.world.getWinningTeam())):
            self.achievementTriggered()


@availableAchievements.register
class RabbitKill(OncePerPlayerPerGame):
    name = 'Icing on the Cake'
    description = 'Kill an enemy after winning a game'
    idstring = 'killRabbit'
    afterGame = True
    inGame = False

    def register(self):
        super(RabbitKill, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(RabbitKill, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, target, *args, **kwargs):
        if (self.world.getWinningTeam() == self.player.team and
                self.player.isEnemyTeam(target.team)):
            self.achievementTriggered()


#disabled @availableAchievements.register
@availableAchievements.register
class TurretKill(OncePerPlayerPerGame):
    name = 'Sentry Mode Activated'
    description = 'Kill an enemy as a turret'
    idstring = 'turretKill'

    def register(self):
        super(TurretKill, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(TurretKill, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, *args, **kwargs):
        if self.player.upgrade and self.player.upgrade.upgradeType == 't':
            self.achievementTriggered()


@availableAchievements.register
class EveryoneDied(OncePerGame):
    idstring = 'everyoneDied'
    name = 'We All Fall Down'
    description = (
        'Be dead at the same time as everyone else in a game of 6 or more '
        'players')

    def register(self):
        super(EveryoneDied, self).register()
        self.world.onPlayerKill.addListener(self.playerKilled)

    def unregister(self):
        super(EveryoneDied, self).unregister()
        self.world.onPlayerKill.removeListener(self.playerKilled)

    def playerKilled(self, killer, *args, **kwargs):
        if len(self.world.players) >= 6:
            if all(p.dead for p in self.world.players):
                self.achievementTriggeredFor(self.world.players)


@availableAchievements.register
class FirstKill(OncePerGame):
    name = 'First Blood'
    description = 'Make the first kill of the game'
    idstring = 'firstKill'

    def register(self):
        super(FirstKill, self).register()
        self.world.onPlayerKill.addListener(self.playerKilled)

    def unregister(self):
        super(FirstKill, self).unregister()
        self.world.onPlayerKill.removeListener(self.playerKilled)

    def playerKilled(self, killer, *args, **kwargs):
        if killer:
            self.achievementTriggeredFor({killer})


@availableAchievements.register
class FirstTag(OncePerGame):
    name = 'Game On'
    description = 'Tag or assist in the tagging of the first zone in a game'
    idstring = 'firstTag'

    def register(self):
        super(FirstTag, self).register()
        self.world.onZoneTagged.addListener(self.zoneTagged)

    def unregister(self):
        super(FirstTag, self).unregister()
        self.world.onZoneTagged.removeListener(self.zoneTagged)

    def zoneTagged(self, zone, player, *args, **kwargs):
        if not player:
            return
        team = player.team
        helpers = {p for p in zone.players if p.team == team and not p.dead}
        self.achievementTriggeredFor(helpers)


@availableAchievements.register
class FirstRespawn(OncePerGame):
    name = 'Finger on the Trigger'
    description = 'Be the first person to respawn (minimum 6 players)'
    idstring = 'firstRespawn'

    def register(self):
        super(FirstRespawn, self).register()
        self.world.onPlayerRespawn.addListener(self.playerRespawned)

    def unregister(self):
        super(FirstRespawn, self).unregister()
        self.world.onPlayerRespawn.removeListener(self.playerRespawned)

    def playerRespawned(self, player, *args, **kwargs):
        if len(self.world.players) < 6:
            self.achievementTriggeredFor([])
        else:
            self.achievementTriggeredFor({player})


@availableAchievements.register
class LastTag(OncePerGame):
    name = 'And the Dirt is Gone'
    description = 'Tag or assist in the tagging of the final zone'
    idstring = 'finalTag'

    def register(self):
        super(LastTag, self).register()
        self.world.onZoneTagged.addListener(self.zoneTagged)

    def unregister(self):
        super(LastTag, self).unregister()
        self.world.onZoneTagged.removeListener(self.zoneTagged)

    def zoneTagged(self, zone, player, *args, **kwargs):
        if not player:
            return
        team = player.team

        if team and team.opposingTeam.numZonesOwned == 0:
            winners = {
                p for p in zone.players
                if p.team == team and not p.dead}
            self.achievementTriggeredFor(winners)


@availableAchievements.register
class LoseWithEntireTeamDead(OncePerGame):
    name = 'Fight to the Death'
    description = 'Lose with entire team dead (minimum 6 players)'
    idstring = 'losingTeamDead'

    def register(self):
        super(LoseWithEntireTeamDead, self).register()
        self.world.onZoneTagged.addListener(self.zoneTagged)

    def unregister(self):
        super(LoseWithEntireTeamDead, self).unregister()
        self.world.onZoneTagged.removeListener(self.zoneTagged)

    def zoneTagged(self, zone, player, *args, **kwarsg):
        if not player:
            return
        if len(self.world.players) < 6:
            return
        team = player.team

        if team and team.opposingTeam.numZonesOwned == 0:
            losingTeam = team.opposingTeam
            losers = {p for p in zone.players if p.team == losingTeam}
            if all(p.dead for p in losers):
                self.achievementTriggeredFor(losers)


@availableAchievements.register
class Domination(OncePerPlayerPerGame):
    name = 'Domination'
    description = 'Kill the same enemy five times in a single game'
    idstring = 'domination'
    streakTarget = 5

    def __init__(self, *args, **kwargs):
        super(Domination, self).__init__(*args, **kwargs)
        self.playerKills = defaultdict(int)

    def register(self):
        super(Domination, self).register()
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(Domination, self).unregister()
        self.player.onKilled.removeListener(self.killedPlayer)

    def killedPlayer(self, target, *args, **kwargs):
        nick = target.identifyingName
        self.playerKills[nick] += 1
        if self.playerKills[nick] == self.streakTarget:
            self.achievementTriggered()


@availableAchievements.register
class MutualKill(NoLimit):
    name = 'Eye for an Eye'
    description = 'Kill an enemy at the same time he kills you'
    idstring = 'mutualKill'

    def __init__(self, *args, **kwargs):
        super(MutualKill, self).__init__(*args, **kwargs)
        # Keep a record of the people we kill. If they kill us while they're
        # dead, achievement unlocked.  Similarly, keep a record of who killed
        # us. If we kill them while we're dead, achievement unlocked.
        self.ourKiller = None
        self.kills = set()

    def register(self):
        super(MutualKill, self).register()
        self.player.onDied.addListener(self.died)
        self.player.onKilled.addListener(self.killedPlayer)
        self.world.onPlayerRespawn.addListener(self.playerRespawned)

    def unregister(self):
        super(MutualKill, self).unregister()
        self.player.onDied.addListener(self.died)
        self.player.onKilled.removeListener(self.killedPlayer)
        self.world.onPlayerRespawn.removeListener(self.playerRespawned)

    def died(self, killer, *args, **kwargs):
        if not killer:
            return
        if killer.identifyingName in self.kills:
            self.achievementTriggered()
        else:
            self.ourKiller = killer.identifyingName
        self.kills = set()

    def killedPlayer(self, target, *args, **kwargs):
        if target.identifyingName == self.ourKiller:
            self.achievementTriggered()
        else:
            self.kills.add(target.identifyingName)
        self.ourKiller = None

    def playerRespawned(self, player, *args, **kwargs):
        self.kills.discard(player.identifyingName)
        if player == self.player:
            self.ourKiller = None


@availableAchievements.register
class SnatchedZone(OncePerPlayerPerGame):
    idstring = 'snatched'
    name = 'Snatched from the Jaws'
    # This should not be awarded if the tagging team would have been able to
    # tag anyway
    description = 'Tag a zone just before an enemy was about to respawn in it'
    snatchTime = 0.5

    def register(self):
        super(SnatchedZone, self).register()
        self.player.onTaggedZone.addListener(self.taggedZone)

    def unregister(self):
        super(SnatchedZone, self).unregister()
        self.player.onTaggedZone.removeListener(self.taggedZone)

    def taggedZone(self, zone, previousOwner, *args, **kwargs):
        if previousOwner is None:
            return

        almostDefenders = 0
        actualDefenders = 0
        actualAttackers = 0
        for player in zone.players:
            if self.player.isFriendsWith(player):
                if not player.dead:
                    actualAttackers += 1
            elif player.team == previousOwner:
                if player.dead:
                    if player.respawnGauge < self.snatchTime:
                        almostDefenders += 1
                else:
                    actualDefenders += 1

        if (not ZoneState.canTag(actualAttackers, actualDefenders +
                almostDefenders)):
            # Couldn't have tagged if we'd been half a second later
            self.achievementTriggered()


@availableAchievements.register
class ShieldRevenge(OncePerPlayerPerGame):
    name = 'Shields Up, Weapons Online'
    description = 'Kill the enemy who shot your shield within ten seconds'
    idstring = 'shieldRevenge'

    timeWindow = 10     # Seconds

    def __init__(self, *args, **kwargs):
        super(ShieldRevenge, self).__init__(*args, **kwargs)
        self.shieldKillers = []

    def register(self):
        super(ShieldRevenge, self).register()
        self.player.onShieldDestroyed.addListener(self.shieldDestroyed)
        self.player.onDied.addListener(self.died)
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(ShieldRevenge, self).unregister()
        self.player.onShieldDestroyed.removeListener(self.shieldDestroyed)
        self.player.onDied.removeListener(self.died)
        self.player.onKilled.removeListener(self.killedPlayer)

    def shieldDestroyed(self, shooter, *args, **kwargs):
        self.shieldKillers.append(
            (self.world.timer.getElapsedTime(), shooter.identifyingName))

    def died(self, *args, **kwargs):
        self.shieldKillers = []

    def killedPlayer(self, target, *args, **kwargs):
        currentTime = self.world.timer.getElapsedTime()
        while self.shieldKillers:
            time, attacker = self.shieldKillers[0]
            if currentTime - time > self.timeWindow:
                self.shieldKillers.pop(0)
            else:
                break

        for time, attacker in self.shieldKillers:
            if attacker == target.identifyingName:
                self.achievementTriggered()
                break


@availableAchievements.register
class ShieldThenKill(OncePerPlayerPerGame):
    name = 'Armour-Piercing Bullets'
    description = 'Kill an enemy after destroying their shields'
    idstring = 'destroyShield'

    timeWindow = 10     # Seconds

    def __init__(self, *args, **kwargs):
        super(ShieldThenKill, self).__init__(*args, **kwargs)
        self.shieldsKilled = []

    def register(self):
        super(ShieldThenKill, self).register()
        self.player.onDestroyedShield.addListener(self.destroyedShield)
        self.player.onDied.addListener(self.died)
        self.player.onKilled.addListener(self.killedPlayer)

    def unregister(self):
        super(ShieldThenKill, self).unregister()
        self.player.onDestroyedShield.removeListener(self.destroyedShield)
        self.player.onDied.removeListener(self.died)
        self.player.onKilled.removeListener(self.killedPlayer)

    def destroyedShield(self, target, *args, **kwargs):
        self.shieldsKilled.append(
            (self.world.timer.getElapsedTime(), target.id))

    def died(self, *args, **kwargs):
        self.shieldsKilled = []

    def killedPlayer(self, target, *args, **kwargs):
        currentTime = self.world.timer.getElapsedTime()
        while self.shieldsKilled:
            time, attacker = self.shieldsKilled[0]
            if currentTime - time > self.timeWindow:
                self.shieldsKilled.pop(0)
            else:
                break

        for time, victim in self.shieldsKilled:
            if victim == target.id:
                self.achievementTriggered()
                break


@availableAchievements.register
class ShieldDefence(OncePerPlayerPerGame):
    name = 'Strategic Deployment'
    description = 'Activate a shield just before being shot'
    idstring = 'shieldDefence'

    timeWindow = 0.2    # seconds
    shieldTime = None

    def register(self):
        super(ShieldDefence, self).register()
        self.player.onUsedUpgrade.addListener(self.usedUpgrade)
        self.player.onHitByShot.addListener(self.hitByShot)

    def unregister(self):
        super(ShieldDefence, self).unregister()
        self.player.onUsedUpgrade.removeListener(self.usedUpgrade)
        self.player.onHitByShot.removeListener(self.hitByShot)

    def usedUpgrade(self, upgrade, *args, **kwargs):
        if upgrade.upgradeType == 's':
            self.shieldTime = self.world.timer.getElapsedTime()

    def hitByShot(self, *args, **kwargs):
        if self.shieldTime is not None:
            currentTime = self.world.timer.getElapsedTime()
            if (currentTime - self.shieldTime) <= self.timeWindow:
                self.achievementTriggered()


@availableAchievements.register
class LastManStanding(OncePerTeamPerGame):
    name = 'Never Give Up'
    description = 'Be the last surviving member of your team'
    idstring = 'surviveLast'
    afterGame = True
    inGame = True

    def register(self):
        super(LastManStanding, self).register()
        self.world.onPlayerKill.addListener(self.playerKilled)
        self.world.onGameOver.addListener(self.gameOver)

    def unregister(self):
        super(LastManStanding, self).unregister()
        self.world.onPlayerKill.removeListener(self.playerKilled)
        self.world.onGameOver.removeListener(self.gameOver)

    def playerKilled(self, killer, *args, **kwargs):
        if not self.world.gameIsInProgress():
            self.check()

    def gameOver(self, team, *args, **kwarsg):
        self.check()

    def check(self):
        remaining = None

        for player in self.world.players:
            if player.team == self.team and not player.dead:
                if remaining is None:
                    remaining = player
                else:
                    return
        if remaining:
            self.achievementTriggeredFor([remaining])


##################
# Leader Killing #
##################
leaderAchievements = {
                      'adonal': 'Drum Solo',
                      'dcatch': 'Jerakeen Genocide',
                      'dmenge': "Trolling Ain't Easy",
                      'jbartl': 'The Bigger They Are...',
                      'jowen':  'Zergling Rush',
                      'jredmo': 'Kernel Panic',
                      'khethe': 'Vespene Geyser Exhausted',
                      'kjohns': 'All Your Bass are Belong to Us',
                      'lsharl': 'Knight of Catan',
                      'mbrand': 'How do I shot web?',
                      'mwalla': "School's Out",
                      'shorn':  'Employee of the Month',
                      'tirwin': 'Uncomfortably Energetic',
                      'tryan':  'Nerf Warfare'
                      }

leaderNames =        {'adonal': 'ErbaneOrb',
                      'dcatch': 'Kindred',
                      'dmenge': 'Cricket',
                      'jbartl': 'Hrodga',
                      'jowen':  'Nobody',
                      'jredmo': 'Avatar',
                      'khethe': 'Scarecrow',
                      'kjohns': 'Kain',
                      'lsharl': 'Dr. Strangelove',
                      'mbrand': 'PorridgeLord',
                      'mwalla': 'Jupator',
                      'shorn':  'Clamburger',
                      'tirwin': 'Bundy',
                      'tryan':  'Acne'
                      }

leaderAchievements = additionalLeaders = {} # We are not on camp.


def getLeaderDescription(name):
    return 'Kill %s three times' % (leaderNames[name],)


# for username, achievementName in leaderAchievements.iteritems():
#     @availableAchievements.register
#     class LeaderMultiKill(Streak, OnceEverPerPlayer):
#         streakTarget = 3
#         idstring = 'leader' + stripPunctuation(username)
#         name = achievementName
#         description = getLeaderDescription(username)
#         username = username
#
#     def register(self):
#         super(TotalKillsBase, self).register()
#         self.player.onKilled.addListener(self.killedPlayer)
#
#     def unregister(self):
#         super(TotalKillsBase, self).unregister()
#         self.player.onKilled.removeListener(self.killedPlayer)
#
#     def killedPlayer(self, *args, **kwargs):
#         def killedSomeone(self, msg):
#             player = self.world.getPlayer(msg.targetId)
#             if player.user is None:
#                 return
#             username = player.user.username
#             if username == self.username:
#                 self.increment()
#
#
# #disabled @availableAchievements.register
# class DeveloperKills(KilledSomeoneAchievement, Streak,
#         OnceEverPerPlayer):
#     idstring = 'developerKills'
#     name = 'Developer Destroyer'
#     description = 'Rack up 10 developer kills'
#     streakTarget = 10
#
#     developers = ('jbartl', 'adonal', 'shorn',
#                   'hrodga', 'erbaneorb', 'clamburger')
#
#     def killedSomeone(self, msg):
#         player = self.world.getPlayer(msg.targetId)
#         if player.user is None:
#             return
#         nick = stripPunctuation(player.user.username).lower()
#         if nick in self.developers:
#             self.increment()

# #disabled @availableAchievements.register
# class KillEveryLeader(KilledSomeoneAchievement, PersistedChecklistAchievement):
#     idstring = 'killEveryLeader'
#     name = 'Leaders vs Campers'
#     description = 'Kill every leader at least once'
#     requiredItems = set(map(stripPunctuation,
#                             map(lower, leaderAchievements.keys()
#                                     + additionalLeaders.keys())))
#
#     def killedSomeone(self, msg):
#         player = self.world.getPlayer(msg.targetId)
#         if player.user is None:
#             return
#         nick = stripPunctuation(player.user.username).lower()
#         if nick in self.requiredItems:
#             self.addItem(nick)



import logging
import uuid

from twisted.web import server, resource
from twisted.internet import defer

log = logging.getLogger(__name__)


class PageSet(object):
    def __init__(self):
        self.pages = {}

    def addPage(self, url, allOrigins=False):
        def pagefn(fn):
            fn._allOrigins = allOrigins
            self.pages[url] = fn
            return fn
        return pagefn


class ServerState(object):
    def __init__(self, pageSet):
        self.uuid = uuid.uuid4()
        self.pageSet = pageSet
        self.site = self._createSite()

    def _createSite(self):
        resources = Resources(self)
        return server.Site(resources)


class Resources(resource.Resource):
    isLeaf = True

    def __init__(self, state):
        resource.Resource.__init__(self)
        self.state = state
        self.pageSet = state.pageSet

    def render_POST(self, request):
        return self.render_GET(request)

    def render_GET(self, request):
        try:
            page = self.pageSet.pages['/'.join(request.postpath)]
        except KeyError:
            request.setResponseCode(404)
            return '<html><body><h1>Not found</h1></body></html>'

        if hasattr(page, '_allOrigins') and page._allOrigins:
            request.setHeader('Access-Control-Allow-Origin', '*')

        d = page(self.state, request)
        if isinstance(d, defer.Deferred):
            d.addCallback(self._requestComplete, request)
            return server.NOT_DONE_YET
        d = self._setRequestCharset(request, d)
        return d

    def _requestComplete(self, result, request):
        if result:
            result = self._setRequestCharset(request, result)
            request.write(result)
        request.finish()

    def _setRequestCharset(self, request, contents):
        if isinstance(contents, str):
            return contents
        assert isinstance(contents, unicode)

        for value in (request.responseHeaders.getRawHeaders('Content-Type') or ['text/plain']):
            if value.startswith('text/'):
                break
        else:
            raise TypeError('Cannot set encoding of non-text document')
        request.responseHeaders.addRawHeader('Content-Type', 'charset=utf-8')
        contents = contents.encode('utf-8')
        return contents

class Template(object):
    def __init__(self, string, defaults):
        self.string = string
        self.defaults = defaults

    def apply(self, **args):
        d = dict(self.defaults)
        d.update(args)
        return self.string % d



'''
Provides extra debugging about which reactor calls are taking a long time.

Note that this adds the overhead of two system clock calls per function call
that the reactor makes.
'''
from contextlib import contextmanager
import cProfile
import logging
import pstats
from StringIO import StringIO

from twisted.internet import base, defer, reactor
from twisted.python import reflect

log = logging.getLogger(__name__)

try:
    from twisted.internet.iocpreactor import reactor as iocpreactor
except ImportError:
    iocpreactor = None

try:
    from twisted.internet import epollreactor
except ImportError:
    epollreactor = None

try:
    from twisted.internet import selectreactor
except ImportError:
    selectreactor = None


profile_everything = False


class FunctionTimer(object):
    '''
    Used to time functions and log those which take longer than a particular
    threshold. If an inner function is also being logged, this is not counted
    towards the outer function's total.

    This is useful for example when deferreds are called back - any queued
    callbacks are called immediately, within the d.callback() call, but if
    they're wrapped it won't be counted towards the caller's total.
    '''
    threshold = 0.1

    def __init__(self):
        self.call_stack = []
        self.message = None
        self.started = 0

    @contextmanager
    def activate(self, message):
        now = reactor.seconds()
        self.call_stack.append((now - self.started, self.message))
        self.started = now
        self.message = message

        pr = None
        if profile_everything:
            pr = cProfile.Profile()
            pr.enable()

        try:
            yield
        finally:
            s = None
            if pr:
                pr.disable()
                s = StringIO()
                ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
                ps.print_stats(40)

            now = reactor.seconds()
            t = now - self.started
            self.log_activity(t, self.message, s)
            t, self.message = self.call_stack.pop()
            self.started = now - t

    def log_activity(self, t, message, s):
        if t > self.threshold:
            if s:
                message += '\n' + s.getvalue()
            log.warning('%.2fs: %s', t, message)

    def wrap(self, fn):
        '''
        Wraps the function so that when it is called, its execution will be
        timed with this timer.
        '''
        def wrapper(*args, **kwargs):
            arg_strings = [repr(arg) for arg in args]
            arg_strings += ['%s=%r' % (k, v) for (k, v) in kwargs.iteritems()]
            name = self.get_name(fn)
            message = ('calling %s(%s)' % (name, ', '.join(arg_strings)))
            with self.activate(message):
                return fn(*args, **kwargs)
        return wrapper

    def get_name(self, fn):
        try:
            return reflect.fullyQualifiedName(fn)
        except:
            return str(fn)

function_timer = FunctionTimer()


def timed_add_callbacks(self, callback, errback=None, *args, **kwargs):
    '''
    Replaces Deferred.addCallbacks so that callbacks are timed.
    '''
    callback = function_timer.wrap(callback)
    if errback is not None:
        errback = function_timer.wrap(errback)
    return monkey_patcher.deferred_addCallbacks(
        self, callback, errback, *args, **kwargs)


def timed_delayed_call(self, time, func, *args, **kwargs):
    '''
    Replaces DelayedCall.__init__ so that delayed calls are timed.
    '''
    func = function_timer.wrap(func)
    return monkey_patcher.delayedCall_init(self, time, func, *args, **kwargs)


class MonkeyPatcher(object):
    def __init__(self):
        self.patched = False
        self.deferred_addCallbacks = None
        self.delayedCall_init = None
        self.ePollReactor_doReadOrWrite = None

    def on(self):
        if self.patched:
            return
        self.deferred_addCallbacks = defer.Deferred.addCallbacks.im_func
        self.delayedCall_init = base.DelayedCall.__init__.im_func
        if epollreactor:
            self.ePollReactor_doReadOrWrite = (
                epollreactor.EPollReactor._doReadOrWrite.im_func)
        if selectreactor:
            self.selectReactor_doReadOrWrite = (
                selectreactor.SelectReactor._doReadOrWrite.im_func)
        if iocpreactor:
            self.iocpReactor_callEventCallback = (
                iocpreactor.IOCPReactor._callEventCallback.im_func)
        self.patched = True
        defer.Deferred.addCallbacks = timed_add_callbacks
        base.DelayedCall.__init__ = timed_delayed_call
        if epollreactor:
            epollreactor.EPollReactor._doReadOrWrite = function_timer.wrap(
                self.ePollReactor_doReadOrWrite)
        if selectreactor:
            selectreactor.SelectReactor._doReadOrWrite = function_timer.wrap(
                self.selectReactor_doReadOrWrite)
        if iocpreactor:
            iocpreactor.IOCPReactor._callEventCallback = function_timer.wrap(
                self.iocpReactor_callEventCallback)

    def off(self):
        if not self.patched:
            return
        defer.Deferred.addCallbacks = self.deferred_addCallbacks
        base.DelayedCall.__init__ = self.delayedCall_init
        if epollreactor:
            epollreactor.EPollReactor._doReadOrWrite = (
                self.ePollReactor_doReadOrWrite)
        if selectreactor:
            selectreactor.SelectReactor._doReadOrWrite = (
                self.selectReactor_doReadOrWrite)
        if iocpreactor:
            iocpreactor.IOCPReactor._callEventCallback = (
                self.iocpReactor_callEventCallback)
        self.patched = False

monkey_patcher = MonkeyPatcher()


def start(profiling=False):
    global profile_everything
    profile_everything = profiling
    monkey_patcher.on()


def stop():
    monkey_patcher.off()


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.DEBUG)
    import time
    monkey_patcher.on()

    def wait(*args, **kwargs):
        time.sleep(.6)
        return 7
    reactor.callLater(.2, wait, 'X')
    d = defer.succeed(4)
    d.addCallback(wait, 'Y')

    reactor.run()



import logging

import pygame

from trosnoth.gui.common import defaultAnchor, addPositions
from trosnoth.gui.framework.framework import Element

log = logging.getLogger(__name__)


class Clock(Element):
    '''
    A graphical clock class; could be used to show a countdown or a count-up
    of the game in progress. Counts to the second.
    '''

    def __init__(self, app, timer, pos, font,
                 colour = (0,0,0), bgColour = None):
        super(Clock, self).__init__(app)
        self.setTimer(timer)
        self.pos = pos
        self.size = None
        self.colour = colour
        self.bgColour = bgColour

        self.font = font

    def setTimer(self, timer):
        self.timer = timer

    # Rather than the size of this object changing as the numbers change,
    # set a fixed size.
    # For no fixed size, call with size as None
    def setFixedSize(self, size):
        self.size = size

    # Since 'None' as a bgColour means 'no background', call this function
    # with bgColour = False to set it to None.
    def setColours(self, fontColour = None, bgColour = None):
        if fontColour is not None:
            self.colour = fontColour
        if bgColour is not None:
            if bgColour is False:
                self.bgColour = None
            else:
                self.bgColour = bgColour

    def tick(self, deltaT):
        self.timer.tick(deltaT)

    def _getSize(self):
        if self.size is None:
            return addPositions(self.font.size(self.app,
                    self.timer.getTimeString()), (2,0))
        else:
            return self.size.getSize(self.app)

    def _getRect(self):
        rect = pygame.Rect((0,0), self._getSize())
        if hasattr(self.pos, 'apply'):
            self.pos.apply(self.app, rect)
        else:
            setattr(rect, defaultAnchor, self.pos)
        return rect

    def _getPt(self):
        return self._getRect().topleft

    def draw(self, surface):
        if self.bgColour is not None:
            clockImage = self.font.render(self.app, self.timer.getTimeString(),
                    True, self.colour, self.bgColour)
            rect = self._getRect()
            surface.fill(self.bgColour, rect)
            clockRect = clockImage.get_rect()
            clockRect.center = rect.center
            surface.blit(clockImage, clockRect.topleft)
        else:
            clockImage = self.font.render(self.app, self.timer.getTimeString(),
                    True, self.colour)
            surface.blit(clockImage, self._getPt())



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.framework.framework import Element, CompoundElement
from trosnoth.gui.framework.hotkey import Hotkey
from trosnoth.gui.framework.tab import Tab
from trosnoth.gui.common import Location, AttachedPoint, ScaledScalar
from trosnoth.utils.event import Event
import pygame

class TabHeader(Element):
    def __init__(self, app, tab, tabContainer, font):
        super(TabHeader, self).__init__(app)
        self.tab = tab
        self.tabContainer = tabContainer
        self.pos = None
        self.font = font

        self._mouseOver = False

    def setDimensions(self, pos, size):
        self.pos = pos
        self.size = size
        self._redraw()

    def _getColour(self):
        if self.hasFocus:
            return self.tabContainer.borderColour
        elif self._mouseOver:
            return (192, 192, 192)
        else:
            return (128, 128, 128)


    def _redraw(self):
        assert self.pos
        self.surface = pygame.surface.Surface(self.size)
        newSurface = self.font.render(self.app, self.tab.caption, True,
                self._getColour())
        newRect = newSurface.get_rect(centerx=self.size[0] / 2)

        x = self.tabContainer._getBorderWidth() + 2
        self.surface.blit(newSurface, (max(x, newRect.left), x))


    def gotFocus(self):
        self._redraw()

    def lostFocus(self):
        self._redraw()

    def processEvent(self, event):
        if (not self.hasFocus and event.type == pygame.MOUSEBUTTONDOWN
                and event.button == 1 and pygame.Rect(self.pos,
                self.size).collidepoint(event.pos)):
            self.tabContainer._tabSelected(self.tab)
        else:
            return event

    def draw(self, screen):
        screen.blit(self.surface, self.pos)
        pygame.draw.rect(screen, self._getColour(), pygame.Rect(self.pos,
                self.size), self.tabContainer._getBorderWidth())

    def tick(self, deltaT):
        if pygame.Rect(self.pos, self.size).collidepoint(
                pygame.mouse.get_pos()):
            if not self._mouseOver:
                self._mouseOver = True
                self._redraw()
        else:
            if self._mouseOver:
                self._mouseOver = False
                self._redraw()

class TabHeaderArrow(Element):
    def __init__(self, app, tabContainer, text, font, onclick, clickable):
        super(TabHeaderArrow, self).__init__(app)
        self.pos = None
        self.tabContainer = tabContainer
        self.font = font
        self.text = text
        self.onclick = onclick
        self.clickable = clickable

        self._mouseOver = False

    def setDimensions(self, pos, size):
        self.pos = pos
        self.size = size
        self._redraw()

    def _getColour(self):
        if not self.clickable():
            return (128, 128, 128)
        if self._mouseOver:
            return (255, 255, 255)
        else:
            return (192, 192, 192)

    def _getSize(self):
        return self.size.getSize(self.app)

    def _getRect(self):
        r = pygame.Rect((0,0), self._getSize())
        if hasattr(self.pos, 'apply'):
            self.pos.apply(self.app, r)
        return r

    def _getPt(self):
        return self._getRect().topleft

    def _redraw(self):
        assert self.pos
        self.surface = pygame.surface.Surface(self._getSize())
        newSurface = self.font.render(self.app, self.text, True,
                self._getColour())
        newRect = newSurface.get_rect(centerx=self._getSize()[0] / 2)

        x = self.tabContainer._getBorderWidth() + 2
        self.surface.blit(newSurface, (max(x, newRect.left), x))

    def gotFocus(self):
        self._redraw()

    def lostFocus(self):
        self._redraw()

    def processEvent(self, event):
        if (self.clickable() and not self.hasFocus and event.type ==
                pygame.MOUSEBUTTONDOWN and
                self._getRect().collidepoint(event.pos)):
            self.onclick()
        else:
            return event

    def draw(self, screen):
        self._redraw()
        screen.blit(self.surface, self._getPt())
        pygame.draw.rect(screen, self._getColour(), self._getRect(),
                self.tabContainer._getBorderWidth())

    def tick(self, deltaT):
        if self._getRect().collidepoint(pygame.mouse.get_pos()):
            if not self._mouseOver:
                self._mouseOver = True
        else:
            if self._mouseOver:
                self._mouseOver = False

class TabHeaderRow(CompoundElement):
    def __init__(self, app, tabContainer):
        super(CompoundElement, self).__init__(app)

        self.firstVisible = 0
        self.skipOnMove = 3
        self.numToDisplay = 0
        self.arrowWidth = arrowWidth = 30

        self.elements = []
        self.headers = []
        self.selectedIndex = -1

        onLeft = lambda: self._moveDirection(True)
        onRight = lambda: self._moveDirection(False)
        canClickLeft = lambda: self.firstVisible > 0
        canClickRight = lambda: (self.firstVisible + self.numToDisplay <
                len(self.headers))
        self.tabContainer = tabContainer

        self._getHeight = _getHeight = tabContainer._getHeaderHeight

        class ArrowSize(object):
            def __init__(self, text):
                self.text = text

            def getSize(self, app):
                return (arrowWidth, _getHeight())

            def __repr__(self):
                return "Arrow Size"

        self.leftArrow = TabHeaderArrow(app, tabContainer, "<",
                tabContainer.font, onLeft, canClickLeft)
        leftLocation = Location(AttachedPoint((0,0), tabContainer._getRect,
                'topleft'), 'topleft')
        leftSize = ArrowSize("<")
        self.leftArrow.setDimensions(leftLocation, leftSize)

        self.rightArrow = TabHeaderArrow(app, tabContainer, ">",
                tabContainer.font, onRight, canClickRight)
        rightLocation = Location(AttachedPoint((0,0), tabContainer._getRect,
                'topright'), 'topright')
        rightSize = ArrowSize(">")
        self.rightArrow.setDimensions(rightLocation, rightSize)

        self._getPt = self.tabContainer._getPt

    def _getMaxHeaderWidth(self):
        return max(self._getHeight() * 5, self._getWidth() / 3)

    def _getMinHeaderWidth(self):
        return self._getHeight()

    def _getWidth(self):
        return self.tabContainer._getSize()[0]


    def __len__(self):
        return len(self.headers)

    def _moveDirection(self, left):
        if left:
            self.firstVisible = max(self.firstVisible - self.skipOnMove, 0)
        else:
            self.firstVisible = min(self.firstVisible + self.skipOnMove,
                    len(self.headers)-self.numToDisplay)

        self._update()

    def newTabHeader(self, header):
        self.headers.append(header)
        self._update()
        self._orderElements()

    def delTabHeader(self, index):
        # Handling the selectedIndex is done in the tabContainer itself
        del self.headers[index]
        if self.selectedIndex == index:
            self.selectedIndex = -1
        self._update()

    def tabSelected(self, index):
        if self.selectedIndex > -1 and self.selectedIndex < len(self.headers):
            self.headers[self.selectedIndex].lostFocus()
            self.headers[self.selectedIndex].hasFocus = False
        self.selectedIndex = index
        self._setFirstVisible()
        self._update()
        self.setFocus(self.headers[self.selectedIndex])

    def _orderElements(self):
        if self.firstVisible + self.numToDisplay > len(self.headers):
            self.firstVisible = len(self.headers) - self.numToDisplay
        if self.selectedIndex == -1:
            self.elements = []
        elif (self.firstVisible <= self.selectedIndex < self.firstVisible +
                self.numToDisplay):
            self.elements = (
                self.headers[self.firstVisible:self.selectedIndex] +
                self.headers[self.selectedIndex + 1:
                    self.firstVisible + self.numToDisplay] +
                [self.headers[self.selectedIndex]])
        else:
            self.elements = self.headers[self.firstVisible:
                    self.firstVisible + self.numToDisplay]
        if self.numToDisplay != len(self.headers):
            self.elements.append(self.leftArrow)
            self.elements.append(self.rightArrow)

    def _setFirstVisible(self):
        if (self.firstVisible <= self.selectedIndex < self.firstVisible +
                self.numToDisplay):
            pass
        elif self.selectedIndex < self.firstVisible:
            self.firstVisible = self.selectedIndex
        elif self.selectedIndex >= self.firstVisible + self.numToDisplay:
            self.firstVisible = self.selectedIndex - self.numToDisplay + 1
        self._orderElements()

    def tick(self, deltaT):
        self._update()
        self._orderElements()
        super(TabHeaderRow, self).tick(deltaT)


    def _update(self):
        if len(self.headers) > 0:
            # m is the size that each tab header would be
            # if we were to fit them all in
            m = self._getWidth() / len(self.headers)

            # Too many headers, so use arrows to navigate
            if m < self._getMinHeaderWidth():
                availableWidth = self._getWidth() - 2 * self.arrowWidth
                self.numToDisplay = availableWidth / self._getMinHeaderWidth()
                width = availableWidth / self.numToDisplay
                remainder = availableWidth % self.numToDisplay
                self._orderElements()
                currentX = self._getPt()[0] + self.arrowWidth

            # Headers all fit, but will be compressed
            elif m < self._getMaxHeaderWidth():
                width = m
                remainder = self._getWidth() % len(self.headers)
                self.numToDisplay = len(self.headers)
                self.firstVisible = 0
                currentX = self._getPt()[0]

            # There are less headers than can stretch across the tab header area
            else:
                width = self._getMaxHeaderWidth()
                remainder = 0
                self.numToDisplay = len(self.headers)
                self.firstVisible = 0
                currentX = self._getPt()[0]

            for x in range (self.firstVisible,
                    self.firstVisible+self.numToDisplay):
                header = self.headers[x]
                thisWidth = width
                # Do a bit of balancing for tab header widths
                if x - self.firstVisible < remainder:
                    thisWidth += 1
                header.setDimensions((currentX, self._getPt()[1]), (thisWidth,
                        self._getHeight()))
                currentX += thisWidth

class TabContainer(CompoundElement):
    def __init__(self, app, area, font, borderColour=(0, 0, 0)):
        super(TabContainer, self).__init__(app)

        self.onTabSelected = Event()

        # area is that of the entire space including tab headers
        self.area = area
        self.font = font

        self._borderWidth = ScaledScalar(3)
        self.borderColour = borderColour

        nextTab = lambda: self._tabSelected((self.selectedIndex + 1) %
                len(self.tabs))
        lastTab = lambda: self._tabSelected((self.selectedIndex - 1) %
                len(self.tabs))
        self.nextTabKey = Hotkey(app, pygame.K_TAB, pygame.KMOD_CTRL)
        self.nextTabKey.onTriggered.addListener(nextTab)
        self.lastTabKey = Hotkey(app, pygame.K_TAB,
                pygame.KMOD_CTRL|pygame.KMOD_SHIFT)
        self.lastTabKey.onTriggered.addListener(lastTab)

        self.tabs = []
        self.tabHeaders = TabHeaderRow(app, self)
        self.selectedIndex = -1

        self.tabHeaders._update()

        self.elements = [self.tabHeaders, self.lastTabKey, self.nextTabKey]
        self.tabHeaders._update()

    def _getBorderWidth(self):
        if hasattr(self._borderWidth, 'getVal'):
            return int(self._borderWidth.getVal(self.app))
        else:
            return self._borderWidth

    def _getHeaderHeight(self):
        linesize = self.font.getLineSize(self.app)
        return int(linesize * 1.1)


    ##
    # Full rect (including headers)
    def _getRect(self):
        return self.area.getRect(self.app)

    def _getSize(self):
        return self._getRect().size

    def _getPt(self):
        return self._getRect().topleft

    ##
    # For the internal rect (not including headers)
    def _getTabPos(self):
        pos = self._getPt()
        return (pos[0], pos[1]+self._getHeaderHeight())

    def _getTabSize(self):
        size = self._getSize()
        return (size[0], size[1]-self._getHeaderHeight())

    def _getTabRect(self):
        return pygame.Rect(self._getTabPos(), self._getTabSize())

    def _getTabInternalRect(self):
        '''
        Returns the pygame.Rect of area which the tab should draw on. (When the
        tab draws it is relative to the top-left of the tab container.)
        '''
        return pygame.Rect((0, self._getHeaderHeight()), self._getTabSize())

    def selectTab(self, index):
        self._tabSelected(index)

    def _tabSelected(self, tab):
        if self.selectedIndex != -1:
            if not self.tabs[self.selectedIndex].beforeDeactivating():
                return
            self.tabs[self.selectedIndex].deactivated()
        if isinstance(tab, Tab):
            assert tab in self.tabs, "This tab must be in our list"
            self.selectedIndex = self.tabs.index(tab)
        else:
            # By index
            self.selectedIndex = tab

        # Notify those who care:
        self.tabHeaders.tabSelected(self.selectedIndex)
        self.onTabSelected.execute(self.selectedIndex)
        self.tabs[self.selectedIndex].activated()

    def addTab(self, tab):
        self.tabs.append(tab)
        tab.container = self

        newHeader = TabHeader(self.app, tab, self, self.font)
        self.tabHeaders.newTabHeader(newHeader)

        self.tabHeaders._update()
        # If it's the only tab, select it
        if self.selectedIndex == -1:
            self._tabSelected(tab)

    def renameTab(self, newName, tab):
        tab.caption = newName
        self.tabHeaders._update()

    def removeTabAt(self, index):
        tab = self.tabs.pop(index)
        tab.container = None

        self.tabHeaders.delTabHeader(index)
        assert len(self.tabs) == len(self.tabHeaders)
        assert len(self.tabs) != 0
        if index == self.selectedIndex:
            self.selectedIndex = -1
            self._tabSelected(min(index, len(self.tabs)-1))
        self.tabHeaders._update()

    def removeTab(self, tab):
        tab.container = None
        index = self.tabs.index(tab)
        self.removeTabAt(index)


    def draw(self, screen):
        super(TabContainer, self).draw(screen)
        assert len(self.tabs) > 0, ('Need to have at least one tab before '
                'drawing a tabContainer')
        self.tabs[self.selectedIndex].draw(screen)
        # Draw a border
        pygame.draw.rect(screen, self.borderColour, self._getTabRect(),
                self._getBorderWidth())

    def tick(self, deltaT):
        super(TabContainer, self).tick(deltaT)
        for i in range(0, len(self.tabs)):
            # tick all the tabs
            self.tabs[i].tick(deltaT)

    def processEvent(self, event):
        event = super(TabContainer, self).processEvent(event)
        if event:
            event = self.tabs[self.selectedIndex].processEvent(event)
        return event

class TabSize(object):
    def __init__(self, tc):
        self.fn = tc._getTabSize
    def getSize(self, app):
        return self.fn()



import logging
import pygame

from trosnoth.gui.framework.elements import TextElement
from trosnoth.gui.framework import clock, timer
from trosnoth.gui.common import (
    Area, FullScreenAttachedPoint, Location, ScaledSize,
)
import trosnoth.gui.framework.framework as framework
from trosnoth.model import gameStates
from trosnoth.utils.twist import WeakCallLater

log = logging.getLogger(__name__)


class GameTimer(framework.CompoundElement):
    def __init__(self, app, game):
        super(GameTimer, self).__init__(app)
        self.world = game.world
        self.app = app

        # Change these constants to say where the box goes
        self.area = Area(
            FullScreenAttachedPoint(ScaledSize(0, -3), 'midtop'),
            ScaledSize(110, 35),
            'midtop')

        self.lineWidth = max(int(3*self.app.screenManager.scaleFactor), 1)
        # Anything more than width 2 looks way too thick
        if self.lineWidth > 2:
            self.lineWidth = 2
        self.notStarted = TextElement(
            self.app, "--:--", self.app.screenManager.fonts.timerFont,
            Location(
                FullScreenAttachedPoint(ScaledSize(0, -9), 'midtop'),
                'midtop'),
            app.theme.colours.timerFontColour)
        self.elements = [self.notStarted]
        self.running = False

        self.gameTimer = None
        self.gameClock = None
        self.timerAdjustLoop = None
        self.mode = 'none'

        # Seconds for two flashes
        self.flashCycle = 0.5
        # Value the countdown has to get to before it starts to flash
        self.flashValue = 30

        self.loop()

    def loop(self):
        self.timerAdjustLoop = WeakCallLater(10, self, 'loop')
        self.syncTimer()

    def checkState(self):
        if self.mode != self.getWorldMode():
            self.mode = self.getWorldMode()

            if self.running and self.mode == 'stopped':
                self.stopTimer()
            elif (not self.running) and self.mode != 'stopped':
                self.startTimer()
            else:
                self.resetTimer()

    def syncTimer(self):
        if self.gameTimer is None:
            return

        if self.mode == 'starting':
            self.gameTimer.counted = self.gameStartCountdownClock()
            self.gameTimer.countTo = self.world.timer.getStartCountdownMax()
        else:
            self.gameTimer.counted = self.world.timer.getElapsedTime()
            if self.world.timer.hasTimeLimit():
                self.gameTimer.countTo = self.world.timer.getMatchDuration()

    def _flash(self, flashState):
        if flashState == 0:
            self.gameClock.setColours(self.app.theme.colours.timerFlashColour)
        else:
            self.gameClock.setColours(self.app.theme.colours.timerFontColour)

    def kill(self):
        if (self.world.state == gameStates.InProgress and
                self.timerAdjustLoop is not None):
            self.timerAdjustLoop.cancel()
        self.gameTimer = None

    def startTimer(self):
        self.running = True
        self.gameClock = clock.Clock(
            self.app,
            None,   # We'll set the timer soon...
            Location(FullScreenAttachedPoint(
                ScaledSize(0, 0), 'midtop'), 'midtop'),
            self.app.screenManager.fonts.timerFont,
            self.app.theme.colours.timerFontColour)

        self.resetTimer()
        self.elements = [self.gameClock]

    def getWorldMode(self):
        if self.world.state.isStartingCountdown():
            return 'starting'
        if self.world.state.getGameTimeMode() != 'running':
            return 'stopped'
        if self.world.timer.hasTimeLimit():
            return 'countdown'
        return 'countup'

    def gameStartCountdownClock(self):
        t = self.world.timer
        # We subtract 0.999 so that the timer is rounding up the seconds rather
        # than rounding them down. This is so that the instant it hits zero,
        # the game starts.
        return t.getStartCountdownMax() - t.getStartCountdown() - 0.999

    def resetTimer(self):
        self.mode = self.getWorldMode()
        if self.mode == 'starting':
            self.gameTimer = timer.CountdownTimer(
                highest='minutes', clock=self.gameStartCountdownClock)
        elif self.mode == 'countdown':
            self.gameTimer = timer.CountdownTimer(
                highest='minutes', clock=self.world.timer.getElapsedTime)
        else:
            self.gameTimer = timer.Timer(
                highest='minutes', clock=self.world.timer.getElapsedTime)
        if self.gameClock is not None:
            self.gameClock.setTimer(self.gameTimer)
            self._flash(1)
        self.syncTimer()
        self.gameTimer.start()

    def stopTimer(self):
        self.running = False
        if self.timerAdjustLoop is not None:
            self.timerAdjustLoop.cancel()
            self.timerAdjustLoop = None
            self.gameTimer.pause()
            if self.gameClock:
                self.gameClock.setColours(
                    self.app.theme.colours.timerFontColour)
            self.gameTimer = None
        self.elements = [self.notStarted]

    def _getRect(self):
        return self.area.getRect(self.app)

    def tick(self, deltaT):
        super(GameTimer, self).tick(deltaT)
        self.checkState()
        if (
                self.running and self.mode == 'countdown'
                and self.gameTimer is not None
                and self.gameTimer.getCurTime() <= self.flashValue):
            self._flash(
                int((self.gameTimer.getCurTime() / self.flashCycle) % 2))

    def draw(self, surface):
        timerBox = self._getRect()
        # Box background
        surface.fill(self.app.theme.colours.timerBackground, timerBox)

        # Box border
        pygame.draw.rect(
            surface, self.app.theme.colours.black, timerBox, self.lineWidth)

        super(GameTimer, self).draw(surface)



import logging

from twisted.internet import defer
from twisted.python.failure import Failure

from trosnoth.utils import netmsg
from trosnoth.network.networkDefines import validServerVersions
from trosnoth.messages import (
    ChatMsg, TaggingZoneMsg, PlayerStarsSpentMsg, PlayerHasUpgradeMsg,
    AbandonUpgradeMsg, CannotBuyUpgradeMsg, GameStartMsg, GameOverMsg,
    StartingSoonMsg, SetTeamNameMsg, SetGameModeMsg, ShotFiredMsg, RespawnMsg,
    PlayerUpdateMsg, AwardPlayerStarMsg, ChatFromServerMsg, AddPlayerMsg,
    SetAgentPlayerMsg, UpdatePlayerStateMsg, AimPlayerAtMsg, RemovePlayerMsg,
    CannotJoinMsg, InitClientMsg, DelayUpdatedMsg, ShotHitPlayerMsg,
    ZoneStateMsg, TickMsg, AchievementUnlockedMsg, WorldResetMsg,
    PlayerIsReadyMsg, PreferredTeamSelectedMsg, SetPreferredDurationMsg,
    SetPreferredSizeMsg, SetPlayerTeamMsg, CreateCollectableStarMsg,
    RemoveCollectableStarMsg, PlayerHasElephantMsg, ChangeNicknameMsg,
    SetGameSpeedMsg, FireShoxwaveMsg, ReturnToLobbyMsg, ChangeTimeLimitMsg,
    UpgradeChangedMsg, PlayerHasTrosballMsg, TrosballPositionMsg,
    TrosballTagMsg, ResyncPlayerMsg, ResyncAcknowledgedMsg, WorldLoadingMsg,
    PlayerNoticedZombieHitMsg, UpgradeApprovedMsg,
)
from trosnoth.network.base import MsgClientProtocol
from trosnoth.utils.event import Event
from trosnoth.utils.unrepr import unrepr

log = logging.getLogger('client')

clientMsgs = netmsg.MessageCollection(
    PlayerHasElephantMsg,
    PlayerHasTrosballMsg,
    InitClientMsg,
    CannotJoinMsg,
    AbandonUpgradeMsg,
    GameStartMsg,
    TaggingZoneMsg,
    PlayerStarsSpentMsg,
    PlayerHasUpgradeMsg,
    ShotFiredMsg,
    AddPlayerMsg,
    SetPlayerTeamMsg,
    RespawnMsg,
    UpdatePlayerStateMsg,
    AimPlayerAtMsg,
    GameOverMsg,
    ReturnToLobbyMsg,
    StartingSoonMsg,
    SetTeamNameMsg,
    SetGameModeMsg,
    SetGameSpeedMsg,
    SetAgentPlayerMsg,
    RemovePlayerMsg,
    PlayerUpdateMsg,
    ResyncPlayerMsg,
    ResyncAcknowledgedMsg,
    CannotBuyUpgradeMsg,
    UpgradeApprovedMsg,
    ChatMsg,
    ZoneStateMsg,
    AchievementUnlockedMsg,
    WorldResetMsg,
    PlayerIsReadyMsg,
    SetPreferredDurationMsg,
    PreferredTeamSelectedMsg,
    SetPreferredSizeMsg,
    CreateCollectableStarMsg,
    RemoveCollectableStarMsg,
    ChangeNicknameMsg,
    FireShoxwaveMsg,
    ChangeTimeLimitMsg,
    UpgradeChangedMsg,
    TrosballPositionMsg,
    TrosballTagMsg,
    PlayerNoticedZombieHitMsg,
    AwardPlayerStarMsg,
    ChatFromServerMsg,
    DelayUpdatedMsg,
    ShotHitPlayerMsg,
    WorldLoadingMsg,
    TickMsg,
)


class ConnectionFailed(Exception):
    def __init__(self, reason):
        self.reason = reason


class TrosnothClientProtocol(MsgClientProtocol):
    # In Trosnoth 1.8, the network protocol changed completely in order to
    # allow the server to send messages to specific agents and not just
    # generally broadcast everything. As a result, the greeting had to change.
    greeting = 'Trosnoth18'
    messages = clientMsgs

    def connectionMade(self):
        self.onConnectionLost = Event()

        super(TrosnothClientProtocol, self).connectionMade()
        self.validated = False
        self.settings_deferred = defer.Deferred()
        self.getSettingsCalled = False

    def getSettings(self):
        assert not self.getSettingsCalled, 'getSettings() already called'
        self.getSettingsCalled = True
        return self.settings_deferred

    def disconnect(self):
        self.transport.loseConnection()

    def gotGeneralMsg(self, msg):
        if not self.validated:
            if isinstance(msg, InitClientMsg):
                self.gotInitClientMsg(msg)
        else:
            super(TrosnothClientProtocol, self).gotGeneralMsg(msg)

    def gotInitClientMsg(self, msg):
        # Settings from the server.
        settings = unrepr(msg.settings)

        # Check that we recognise the server version.
        svrVersion = settings.get('serverVersion', 'server.v1.0.0+')
        if svrVersion not in validServerVersions:
            log.info('Client: bad server version %s', svrVersion)
            self.settings_deferred.errback(Failure(
                ConnectionFailed('Incompatible server version.')))
            self.transport.abortConnection()
            return

        # Tell the client that the connection has been made.
        self.validated = True
        self.settings_deferred.callback(settings)

    def receiveBadString(self, string):
        if not self.validated:
            self.settings_deferred.errback(Failure(ConnectionFailed(
                'Remote host sent unexpected message: %r' % (string,))))
            self.transport.abortConnection()
            return

        log.warning('Client: Unknown message: %r', string,)
        log.warning('      : Did you invent a new network message and forget')
        log.warning('      : to add it to trosnoth.network.client.clientMsgs?')

    def connectionLost(self, reason=None):
        if not self.settings_deferred.called:
            self.settings_deferred.errback(Failure(
                ConnectionFailed('Remote server dropped connection.')))
            return
        super(TrosnothClientProtocol, self).connectionLost(reason)
        self.onConnectionLost()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging

from trosnoth.model.unit import Bouncy, CollectableUnit, PredictedBouncyTrajectory

log = logging.getLogger(__name__)


class Trosball(Bouncy, CollectableUnit):
    # The following values control star movement.
    maxFallVel = 540            # pix/s
    gravity = 1000              # pix/s/s

    HALF_WIDTH = 10
    HALF_HEIGHT = 10

    def __init__(self, world, *args, **kwargs):
        super(Trosball, self).__init__(world, *args, **kwargs)

        layout = world.map.layout
        self.pos = self.oldPos = (layout.centreX, layout.centreY)

        self.xVel = 0
        self.yVel = 0

        self.inNet = False

    def checkCollision(self, player, delay):
        if self.inNet:
            return False
        if self.world.trosballCooldownPlayer == player:
            return False
        return super(Trosball, self).checkCollision(player, delay)

    def collidedWithPlayer(self, player):
        self.world.sendTrosballCollect(player)

    def setIsInNet(self, pos):
        self.pos = self.oldPos = pos
        self.xVel = 0
        self.yVel = 0
        self.inNet = True

    def advance(self):
        if not self.inNet:
            super(Trosball, self).advance()

    def getGravity(self):
        return self.gravity

    def getMaxFallVel(self):
        return self.maxFallVel

    def teleport(self, pos, xVel, yVel):
        self.pos = self.oldPos = pos
        self.xVel = xVel
        self.yVel = yVel
        self.hitLocalPlayer = False

    def continueOffMap(self):
        self.world.resetTrosballToCentreOfMap()
        return False

        
class PredictedTrosballTrajectory(PredictedBouncyTrajectory):
    HALF_WIDTH = Trosball.HALF_WIDTH
    HALF_HEIGHT = Trosball.HALF_HEIGHT
    def __init__(self, world, player):
        PredictedBouncyTrajectory.__init__(self, world, player, 4, world.physics.trosballThrowVel, Trosball.gravity, Trosball.maxFallVel)
        
    def explosionRadius(self):
        return 0